import collections
import collections.abc
from functools import wraps
import typing
from typing import TypeVar, Generic
import weakref


class WeakObjectIdDict(collections.MutableMapping):
    """
    Like weakref.WeakKeyDict but internally uses object ids instead of the object reference
    itself thereby avoiding the need for the object to be hashable (and therefore immutable).
    """

    def __init__(self, seq=None, **kwargs):
        self._refs = {}  # type: collections.abc.MutableMapping[int, weakref.ReferenceType]
        self._values = {}  # type: collections.abc.MutableMapping[int, typing.Any]
        if seq:
            if isinstance(seq, collections.abc.Mapping):
                for key, value in seq.items():
                    self[key] = value
            elif isinstance(seq, collections.Iterable):
                for key, value in seq:
                    self[key] = value
        if kwargs:
            for key, value in kwargs.items():
                self[key] = value

    def __copy__(self):
        return WeakObjectIdDict(self)

    def __getitem__(self, item):
        try:
            return self._values[id(item)]
        except KeyError:
            raise KeyError(repr(item))

    def __setitem__(self, key, value):
        obj_id = id(key)
        wref = weakref.ref(key, self._finalised)
        self._refs[obj_id] = wref
        self._values[obj_id] = value

    def __delitem__(self, key):
        obj_id = id(key)
        del self._values[obj_id]
        del self._refs[obj_id]

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        for ref in self._refs.values():
            yield ref()

    def _finalised(self, wref):
        found_id = None
        for obj_id, ref in self._refs.items():
            if ref == wref:
                found_id = obj_id
                break
        # Delete both the object values and the reference itself
        del self._values[found_id]
        del self._refs[found_id]


T = TypeVar('T')  # Declare type variable pylint: disable=invalid-name


class DefaultFromCall:
    """Can be used as a default that is generated from a callable when needed"""

    def __init__(self, default_fn):
        assert callable(default_fn), "Must supply callable"
        self._callable = default_fn

    def __call__(self, *args, **kwargs):
        return self._callable(*args, **kwargs)


class NamedTupleBuilder(Generic[T]):
    """A builder that allows namedtuples to be build step by step"""

    def __init__(self, tuple_type: typing.Type[T], defaults=None):
        # Have to do it this way because we overwrite __setattr__
        defaults = defaults or {}
        object.__setattr__(self, '_tuple_type', tuple_type)
        diff = set(defaults.keys()) - set(tuple_type._fields)
        assert not diff, "Can't supply defaults that are not in the namedtuple: '{}'".format(diff)
        object.__setattr__(self, '_values', defaults)

    def __repr__(self):
        """Representation of the object."""
        return '%s(%s)' % (self.__class__.__name__, dict.__repr__(self._values))

    def __getattr__(self, attr):
        """Read a key as an attribute.

        :raises AttributeError: if the attribute does not correspond to an existing key.
        """
        try:
            return self._values[attr]
        except KeyError:
            errmsg = "'{}' object has no attribute '{}'".format(self.__class__.__name__, attr)
            raise AttributeError(errmsg)

    def __setattr__(self, attr, value):
        """Set a key as an attribute."""
        if attr not in self._tuple_type._fields:
            raise AttributeError("AttributeError: '{}' is not a valid attribute of the object "
                                 "'{}'".format(attr, self.__class__.__name__))

        self._values[attr] = value

    def __dir__(self):
        return self._tuple_type._fields

    def update(self, new_values: dict):
        for key, value in new_values.items():
            setattr(self, key, value)

    def build(self) -> T:
        build_from = {
            key: value if not isinstance(value, DefaultFromCall) else value()
            for key, value in self._values.items()
        }
        return self._tuple_type(**build_from)


def to_slice(specifier) -> slice:
    """
    Turn the specifier into a slice object.  Accepts either:
    1. a slice, which is just returned.
    2. a concrete index (positive or negative) e.g. to_slice(5) will generate a slice that
        returns the 5th element
    3. a string containing '*' or ':' to get all entries.
    """
    if isinstance(specifier, slice):
        return specifier
    if isinstance(specifier, int):
        sign = -1 if specifier < 0 else 1
        return slice(specifier, specifier + sign, sign)
    if isinstance(specifier, str) and specifier == ':' or specifier == '*':
        return slice(None)

    raise ValueError("Unknown slice specifier: {}".format(specifier))


def sync(save=False):
    """Decorator that will call sync before the method is invoked to make sure the object is up
    to date what what's in the database."""

    def inner(obj_method):

        @wraps(obj_method)
        def wrapper(self, *args, **kwargs):
            # pylint: disable=protected-access
            try:
                self.__sync += 1
            except AttributeError:
                self.__sync = 1
            if self.is_saved():
                self.sync()
            try:
                retval = obj_method(self, *args, **kwargs)
                if save:
                    self.save()
                return retval
            finally:
                self.__sync -= 1

        return wrapper

    return inner



import mincepy
from mincepy.testing import *
from mincepy import builtins

# region lists


def test_ref_list(historian: mincepy.Historian):
    """Test references list"""
    list1 = mincepy.RefList()
    list2 = mincepy.RefList()
    car = Car()
    list1.append(car)
    list2.append(car)

    # Test iteration
    for entry in list1:
        assert entry is car

    # Reference condition satisfied
    assert list1[0] is list2[0]

    # Now delete everything, reload, and make sure the condition is still satisfied
    list1_id, list2_id = historian.save(list1, list2)
    assert car.is_saved(), "The container should automatically save all it's entries saved"
    del list1, list2, car

    list1_loaded = historian.load(list1_id)
    list2_loaded = historian.load(list2_id)
    assert len(list1_loaded) == 1
    assert len(list2_loaded) == 1
    assert isinstance(list1_loaded[0], Car)
    assert list1_loaded[0] is list2_loaded[0]


def test_ref_list_primitives(historian: mincepy.Historian):
    """Test that we can store primitives in a ref list also"""
    reflist = mincepy.RefList()
    reflist.append(5)
    reflist.append('hello')
    reflist.append(10.8)

    reflist_id = reflist.save()
    del reflist

    loaded = historian.load(reflist_id)
    assert loaded[0] == 5
    assert loaded[1] == 'hello'
    assert loaded[2] == 10.8


def test_live_lists(historian: mincepy.Historian):
    """Basic tests on live list. It's difficult to test the 'syncing' aspects
    of this container without another process"""
    for list_type in (builtins.LiveList, builtins.LiveRefList):
        live_list = list_type()

        car = Car()

        live_list.append(car)
        wrapper_id = live_list.save()
        del live_list

        live_list = historian.load(wrapper_id)
        assert len(live_list) == 1

        # Test getitem
        live_list.append(Car('honda'))
        live_list.append(Car('fiat'))
        assert live_list[2].make == 'fiat'

        # Test delete
        del live_list[2]
        assert len(live_list) == 2


def test_ref_list_none(historian: mincepy.Historian):
    """Test that we can store None in a ref list."""
    lrlist = mincepy.RefList()
    lrlist.append(None)
    assert lrlist[-1] is None
    lrdict_id = lrlist.save()
    del lrlist

    loaded = historian.load(lrdict_id)  # type: mincepy.RefList
    assert loaded[-1] is None


# endregion

# region dict


def test_ref_dict(historian: mincepy.Historian):
    """Test that the ref dict correctly stores entries as references"""
    dict1 = mincepy.builtins.RefDict()
    dict2 = mincepy.builtins.RefDict()
    car = Car()
    dict1['car'] = car
    dict2['car'] = car

    # Test iteration
    for key, value in dict1.items():
        assert key == 'car'
        assert value is car

    # Reference condition satisfied
    assert dict1['car'] is dict2['car']

    # Now delete everything, reload, and make sure the condition is still satisfied
    dict1_id, dict2_id = historian.save(dict1, dict2)
    del dict1, dict2, car

    dict1_loaded = historian.load(dict1_id)
    dict2_loaded = historian.load(dict2_id)
    assert len(dict1_loaded) == 1
    assert len(dict2_loaded) == 1
    assert isinstance(dict1_loaded['car'], Car)
    assert dict1_loaded['car'] is dict2_loaded['car']


def test_ref_dict_primitives(historian: mincepy.Historian):
    """Test that we can store primitives in a ref list also"""
    refdict = mincepy.RefDict()
    refdict['0'] = 5
    refdict['1'] = 'hello'
    refdict['2'] = 10.8

    reflist_id = refdict.save()
    del refdict

    loaded = historian.load(reflist_id)
    assert loaded['0'] == 5
    assert loaded['1'] == 'hello'
    assert loaded['2'] == 10.8


def test_live_dicts(historian: mincepy.Historian):
    """Test live dictionary, difficult to test the sync capability"""
    for dict_type in (builtins.LiveDict, builtins.LiveRefDict):
        dict1 = dict_type()
        dict2 = dict_type()
        car = Car('mini', 'green')
        dict1['car'] = car
        dict2['car'] = car

        # Reference condition satisfied
        assert dict1['car'] is dict2['car']

        # Now delete everything, reload, and make sure the condition is still satisfied
        dict1_id, dict2_id = historian.save(dict1, dict2)
        del dict1, dict2, car

        dict1_loaded = historian.load(dict1_id)
        dict2_loaded = historian.load(dict2_id)
        assert len(dict1_loaded) == 1
        assert len(dict2_loaded) == 1
        assert isinstance(dict1_loaded['car'], Car)
        assert dict1_loaded['car'] == dict2_loaded['car']


def test_live_ref_dict(historian: mincepy.Historian):
    dict1 = builtins.LiveRefDict()
    dict2 = builtins.LiveRefDict()
    car = Car('mini', 'green')
    dict1['car'] = car
    dict2['car'] = car

    # Reference condition satisfied
    assert dict1['car'] is dict2['car']

    # Now delete everything, reload, and make sure the condition is still satisfied
    dict1_id, dict2_id = historian.save(dict1, dict2)
    del dict1, dict2, car

    dict1_loaded = historian.load(dict1_id)
    dict2_loaded = historian.load(dict2_id)
    assert len(dict1_loaded) == 1
    assert len(dict2_loaded) == 1
    assert isinstance(dict1_loaded['car'], Car)
    # Check they are the same object
    assert dict1_loaded['car'] is dict2_loaded['car']


def test_ref_dict_none(historian: mincepy.Historian):
    """Test that we can store None in a ref dict."""
    lrdict = mincepy.LiveRefDict()
    lrdict['test'] = None
    assert lrdict['test'] is None
    lrdict_id = lrdict.save()
    del lrdict

    loaded = historian.load(lrdict_id)  # type: mincepy.RefDict
    assert loaded['test'] is None


# endregion


def test_live_ref_create(historian: mincepy.Historian):
    car = Car('mini', 'green')
    lrdict = builtins.LiveRefDict({'car': car})
    lrdict['car'] = car

    # Now delete everything, reload, and make sure the condition is still satisfied
    dict1_id = historian.save(lrdict)
    car_id = lrdict['car'].obj_id
    assert car_id is not None
    del lrdict, car

    rldict_loaded = historian.load(dict1_id)
    assert len(rldict_loaded) == 1
    assert isinstance(rldict_loaded['car'], Car)
    # Check they are the same object
    assert rldict_loaded['car'].obj_id == car_id


def test_live_ref_lists(historian: mincepy.Historian):
    """Basic tests on live list. It's difficult to test the 'syncing' aspects
    of this container without another process"""
    live_list = builtins.LiveRefList()

    car = Car()

    live_list.append(car)
    wrapper_id = live_list.save()
    car_id = car.obj_id
    assert car_id is not None
    del live_list

    live_list = historian.load(wrapper_id)
    assert len(live_list) == 1

    # Test getitem
    live_list.append(Car('honda'))
    live_list.append(Car('fiat'))
    assert live_list[2].make == 'fiat'

    assert car.obj_id == car_id

    # Test delete
    del live_list[2]
    assert len(live_list) == 2


def test_file_from_disk(tmp_path, historian: mincepy.Historian):
    disk_file_path = tmp_path / 'test.txt'
    disk_file_path.write_text('TEST')
    file = historian.create_file('test.txt')
    file.from_disk(disk_file_path)
    contents = file.read_text()
    assert contents == 'TEST'


def test_file_to_disk(tmp_path, historian: mincepy.Historian):
    disk_file_path = tmp_path / 'test.txt'

    file = historian.create_file('test.txt')
    file.write_text('TEST')

    file.to_disk(tmp_path)

    assert disk_file_path.read_text() == 'TEST'



import contextlib
import functools
import uuid

from . import base_savable

__all__ = 'Process', 'track'


def track(func):

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.running():
            return func(self, *args, **kwargs)

    return wrapper


class Process(base_savable.BaseSavableObject):
    TYPE_ID = uuid.UUID('bcf03171-a1f1-49c7-b890-b7f9d9f9e5a2')
    STACK = []
    ATTRS = '_name', '_running'

    @classmethod
    def current_process(cls):
        if not cls.STACK:
            return None
        return cls.STACK[-1]

    def __init__(self, name: str):
        super(Process, self).__init__()
        self._name = name
        self._running = 0

    def __eq__(self, other):
        if not isinstance(other, Process):
            return False

        return self.name == other.name

    @property
    def is_running(self):
        return self._running != 0

    @property
    def name(self) -> str:
        return self._name

    @contextlib.contextmanager
    def running(self):
        self.STACK.append(self)
        self._running += 1
        try:
            yield
        finally:
            if self.STACK[-1] != self:
                raise RuntimeError("Someone has corrupted the process stack!\n"
                                   "Expected to find '{}' on top but found:{}".format(
                                       self, self.STACK))
            self._running -= 1
            self.STACK.pop()



import collections
import typing

from . import types

__all__ = 'BaseSavableObject', 'AsRef'

AsRef = collections.namedtuple('AsRef', 'attr')

AttrSpec = collections.namedtuple('AttrSpec', 'name as_ref')


class BaseSavableObject(types.SavableObject):
    """A helper class that makes a class compatible with the historian by flagging certain
    attributes which will be saved/loaded/hashed and compared in __eq__.  This should be an
    exhaustive list of all the attributes that define this class.  If more complex functionality
    is needed then the standard SavableComparable interface methods should be overwritten."""
    ATTRS = tuple()
    IGNORE_MISSING = True  # When loading ignore attributes that are missing in the record

    def __new__(cls, *_args, **_kwargs):
        new_instance = super(BaseSavableObject, cls).__new__(cls)
        attrs = {}
        for entry in cls.__mro__:
            try:
                for name in getattr(entry, 'ATTRS'):
                    if name not in attrs:
                        if isinstance(name, AsRef):
                            spec = AttrSpec(name.attr, True)
                        else:
                            spec = AttrSpec(name, False)
                        attrs[name] = spec

            except AttributeError:
                pass
        setattr(new_instance, '__attrs', attrs.values())
        return new_instance

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return all(
            getattr(self, attr.name) == getattr(other, attr.name) for attr in self.__get_attrs())

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables([getattr(self, attr.name) for attr in self.__get_attrs()])

    def save_instance_state(self, _saver) -> dict:
        from . import refs

        saved_state = {}
        for attr in self.__get_attrs():
            item = getattr(self, attr.name)
            if attr.as_ref:
                item = refs.ObjRef(item)
            saved_state[attr.name] = item
        return saved_state

    def load_instance_state(self, saved_state, _loader):
        super(BaseSavableObject, self).load_instance_state(saved_state, _loader)
        for attr in self.__get_attrs():
            try:
                obj = saved_state[attr.name]
            except KeyError:
                if not self.IGNORE_MISSING:
                    raise
            else:
                if attr.as_ref:
                    obj = obj()
                setattr(self, attr.name, obj)

    def __get_attrs(self) -> typing.Sequence[AttrSpec]:
        return getattr(self, '__attrs')



from argparse import Namespace
import pathlib
import uuid

from . import helpers

__all__ = 'PathHelper', 'TupleHelper', 'NamespaceHelper'


class PathHelper(helpers.BaseHelper):
    TYPE = pathlib.Path
    TYPE_ID = uuid.UUID('78e5c6b8-f194-41ae-aead-b231953318e1')
    IMMUTABLE = True

    def yield_hashables(self, obj: pathlib.Path, hasher):
        yield from hasher.yield_hashables(str(obj))

    def save_instance_state(self, obj: pathlib.Path, _saver):
        return str(obj)

    def new(self, encoded_saved_state):
        return pathlib.Path(encoded_saved_state)

    def load_instance_state(self, obj: pathlib.Path, saved_state, _loader):
        pass  # Done it all in new


class TupleHelper(helpers.BaseHelper):
    TYPE = tuple
    TYPE_ID = uuid.UUID('fd9d2f50-71d6-4e70-90b7-117f23d9cbaf')
    IMMUTABLE = True

    def save_instance_state(self, obj: tuple, _saver):
        return list(obj)

    def new(self, encoded_saved_state):
        return tuple(encoded_saved_state)

    def load_instance_state(self, obj: pathlib.Path, saved_state, _loader):
        pass  # Done it all in new

    def yield_hashables(self, obj, hasher):
        for entry in obj:
            yield from hasher.yield_hashables(entry)


class NamespaceHelper(helpers.BaseHelper):
    TYPE = Namespace
    TYPE_ID = uuid.UUID('c43f8329-0d68-4d12-9a35-af8f5ecc4f90')

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(vars(obj))

    def save_instance_state(self, obj, _saver):
        return vars(obj)

    def load_instance_state(self, obj, saved_state, _loader):
        obj.__init__(**saved_state)


HISTORIAN_TYPES = NamespaceHelper(), TupleHelper(), PathHelper()



import mincepy

from mincepy.testing import Car


def test_savable_object(historian: mincepy.Historian):
    """Test basic properties and functions of a savable object"""
    car = Car('smart', 'black')
    car_id = car.save()
    car.set_meta({'reg': 'VD395'})
    del car

    loaded = historian.load(car_id)  # type: Car
    assert loaded.make == 'smart'
    assert loaded.colour == 'black'
    assert loaded.get_meta() == {'reg': 'VD395'}

    loaded.update_meta({'driver': 'martin'})
    assert loaded.get_meta() == {'reg': 'VD395', 'driver': 'martin'}

    assert loaded.obj_id == car_id

    loaded.make = 'honda'
    loaded.save()
    del loaded

    reloaded = historian.load(car_id)
    assert reloaded.make == 'honda'



import contextlib
from typing import MutableMapping, Any, List, Sequence
import weakref

import mincepy.records
from . import archives
from . import exceptions
from . import utils


class LiveObjects:

    def __init__(self):
        # Live object -> data records
        self._records = utils.WeakObjectIdDict()  # type: MutableMapping[Any, archives.DataRecord]
        # Obj id -> object
        self._objects = weakref.WeakValueDictionary()  # type: MutableMapping[Any, Any]

    def __str__(self):
        return "{} live".format(len(self._objects))

    def __contains__(self, item):
        """Determine if an object instance is in this live objects container"""
        return item in self._records

    def insert(self, obj, record):
        self._records[obj] = record
        self._objects[record.obj_id] = obj

    def update(self, live_objects):
        """Like a dictionary update, take the given live objects container and absorb it into ourselves
        overwriting any existing values and incorporating any new"""
        self._records.update(live_objects._records)
        self._objects.update(live_objects._objects)

    def delete(self, obj):
        del self._objects[self.get_record(obj).obj_id]
        del self._records[obj]

    def get_record(self, obj):
        try:
            return self._records[obj]
        except KeyError:
            raise exceptions.NotFound("No live object found '{}'".format(obj))

    def get_object(self, obj_id):
        try:
            return self._objects[obj_id]
        except KeyError:
            raise exceptions.NotFound("No live object with id '{}'".format(obj_id))


class RollbackTransaction(Exception):
    pass


class Transaction:

    def __init__(self):
        # Records staged for saving to the archive
        self._staged = []  # type: List[archives.DataRecord]

        self._live_objects = LiveObjects()
        # Ref -> obj
        self._live_object_references = {}

        # Snapshots: ref -> obj
        self._snapshots = {}  # type: MutableMapping[archives.Ref, Any]

    def __str__(self):
        return "{}, {} live ref(s), {} snapshots, {} staged".format(
            self._live_objects, len(self._live_object_references), len(self._snapshots),
            len(self._staged))

    @property
    def live_objects(self) -> LiveObjects:
        return self._live_objects

    @property
    def snapshots(self):
        return self._snapshots

    # region LiveObjects

    def insert_live_object(self, obj, record):
        ref = record.get_reference()
        if ref not in self._live_object_references:
            self._live_object_references[ref] = obj
        else:
            assert self._live_object_references[ref] is obj
        self._live_objects.insert(obj, record)

    def get_live_object(self, obj_id):
        return self._live_objects.get_object(obj_id)

    def get_record_for_live_object(self, obj):
        return self._live_objects.get_record(obj)

    def insert_live_object_reference(self, ref, obj):
        self._live_object_references[ref] = obj

    def get_live_object_from_reference(self, ref):
        try:
            return self._live_object_references[ref]
        except KeyError:
            raise exceptions.NotFound("No live object with reference '{}' found".format(ref))

    def get_reference_for_live_object(self, obj):
        for ref, cached in self._live_object_references.items():
            if obj is cached:
                return ref
        raise exceptions.NotFound("Live object '{} not found".format(obj))

    def delete(self, obj):
        """Delete an object form the transaction"""
        self._live_objects.delete(obj)
        found_ref = None
        for ref, referenced in self._live_object_references.items():
            if referenced is obj:
                found_ref = ref
                break
        del self._live_object_references[found_ref]

    # endregion LiveObjects

    def insert_snapshot(self, obj, ref):
        self._snapshots[ref] = obj

    def get_snapshot(self, ref):
        try:
            return self._snapshots[ref]
        except KeyError:
            raise exceptions.NotFound("No snapshot with reference '{}' found".format(ref))

    def stage(self, record: mincepy.records.DataRecord):
        """Stage a record to be saved once on completion of this transaction"""
        self._staged.append(record)

    @property
    def staged(self) -> Sequence[mincepy.records.DataRecord]:
        """The list of records that were staged during this transaction"""
        return self._staged

    @contextlib.contextmanager
    def nested(self):
        nested = NestedTransaction(self)
        try:
            yield nested
        except RollbackTransaction:
            pass
        else:
            # Update our transaction with the nested
            self._update(nested)

    def _update(self, transaction):
        self._live_objects.update(transaction.live_objects)
        self._snapshots.update(transaction.snapshots)
        self._live_object_references.update(transaction._live_object_references)
        self._staged.extend(transaction.staged)

    @staticmethod
    def rollback():
        raise RollbackTransaction


class NestedTransaction(Transaction):

    def __init__(self, parent: Transaction):
        super(NestedTransaction, self).__init__()
        self._parent = parent

    def __str__(self):
        return "{} (parent: {})".format(super().__str__(), self._parent)

    def get_live_object(self, obj_id):
        try:
            return super().get_live_object(obj_id)
        except exceptions.NotFound:
            return self._parent.get_live_object(obj_id)

    def get_live_object_from_reference(self, ref):
        try:
            return super().get_live_object_from_reference(ref)
        except exceptions.NotFound:
            return self._parent.get_live_object_from_reference(ref)

    def get_reference_for_live_object(self, obj):
        try:
            return super().get_reference_for_live_object(obj)
        except exceptions.NotFound:
            return self._parent.get_reference_for_live_object(obj)

    def get_record_for_live_object(self, obj):
        try:
            return super().get_record_for_live_object(obj)
        except exceptions.NotFound:
            return self._parent.get_record_for_live_object(obj)

    def get_snapshot(self, ref):
        try:
            return super(NestedTransaction, self).get_snapshot(ref)
        except exceptions.NotFound:
            return self._parent.get_snapshot(ref)



from abc import ABCMeta, abstractmethod
import datetime
import typing
import uuid

try:  # Python3
    from hashlib import blake2b
except ImportError:  # Python < 3.6
    from pyblake2 import blake2b

from . import exceptions

__all__ = 'Savable', 'Comparable', 'Object', 'SavableObject', 'PRIMITIVE_TYPES'

# The primitives that all archive types must support
PRIMITIVE_TYPES = (bool, int, float, str, dict, list, type(None), bytes, uuid.UUID,
                   datetime.datetime)


class Savable(metaclass=ABCMeta):
    """Interface for an object that can save an load its instance state"""
    TYPE_ID = None

    def __init__(self):
        assert self.TYPE_ID is not None, "Must set the TYPE_ID for an object to be savable"

    @abstractmethod
    def save_instance_state(self, saver):
        """Save the instance state of an object, should return a saved instance"""

    @abstractmethod
    def load_instance_state(self, saved_state, loader):
        """Take the given object and load the instance state into it"""


class Comparable(metaclass=ABCMeta):
    """Interface for an object that can be compared and hashed"""

    @abstractmethod
    def __eq__(self, other) -> bool:
        """Determine if two objects are equal"""

    @abstractmethod
    def yield_hashables(self, hasher):
        """Produce a hash representing the value"""


class Object(Comparable, metaclass=ABCMeta):
    """A simple object that is comparable"""


class SavableObject(Object, Savable, metaclass=ABCMeta):
    """A class that is both savable and comparable"""

    def __init__(self, historian=None):
        super().__init__()
        from . import history
        # Tell the historian that we've been created
        self._historian = historian or history.get_historian()
        assert self._historian is not None, \
            "Must provide a valid historian or set one globally using mincepy.set_historian()"

        if self._historian is not None:
            self._historian.created(self)

    @property
    def obj_id(self):
        try:
            return self._historian.get_obj_id(self)
        except exceptions.NotFound:
            return None

    def get_meta(self) -> dict:
        """Get the metadata dictionary for this object"""
        return self._historian.get_meta(self)

    def set_meta(self, meta: typing.Optional[dict]):
        """Set the metadata dictionary for this object"""
        self._historian.set_meta(self, meta)

    def update_meta(self, meta: dict):
        """Update the metadata dictionary for this object"""
        self._historian.update_meta(self, meta)

    def is_saved(self) -> bool:
        return self.obj_id is not None

    def save(self, with_meta=None, return_sref=False):
        """Save the object"""
        return self._historian.save(self, with_meta=with_meta, return_sref=return_sref)

    def sync(self):
        """Update the state of this object by loading the latest version from the historian"""
        return self._historian.sync(self)

    def load_instance_state(self, saved_state, loader):
        """Take the given object and load the instance state into it"""
        super().load_instance_state(saved_state, loader)
        self._historian = loader.get_historian()


class Equator:

    def __init__(self, equators=tuple()):
        self._equators = list(equators)

        def do_hash(*args):
            hasher = blake2b(digest_size=32)
            for arg in args:
                hasher.update(arg)

            return hasher.hexdigest()

        self._hasher = do_hash

    def add_equator(self, equator):
        self._equators.append(equator)

    def remove_equator(self, equator):
        self._equators.reverse()
        try:
            self._equators.remove(equator)
        except ValueError:
            raise ValueError("Unknown equator '{}'".format(equator))
        finally:
            self._equators.reverse()

    def get_equator(self, obj):
        # Iterate in reversed order i.e. the latest added should be used preferentially
        for equator in reversed(self._equators):
            if isinstance(obj, equator.TYPE):
                return equator
        raise TypeError("Don't know how to compare '{}' types, no type equator set".format(
            type(obj)))

    def yield_hashables(self, obj):
        try:
            equator = self.get_equator(obj)
        except TypeError:
            # Try the objects's method
            try:
                yield from obj.yield_hashables(self)
            except AttributeError:
                raise TypeError("No helper registered and no yield_hashabled method on '{}'".format(
                    type(obj)))
        else:
            yield from equator.yield_hashables(obj, self)

    def hash(self, obj):
        return self._hasher(*self.yield_hashables(obj))

    def eq(self, obj1, obj2) -> bool:  # pylint: disable=invalid-name
        if not type(obj1) == type(obj2):
            return False

        try:
            equator = self.get_equator(obj1)
        except TypeError:
            # Fallback to python eq
            return obj1 == obj2
        else:
            return equator.eq(obj1, obj2)

    def float_to_str(self, value, sig=14):  # pylint: disable=no-self-use
        """
        Convert float to text string for computing hash.
        Preserve up to N significant number given by sig.

        :param value: the float value to convert
        :param sig: choose how many digits after the comma should be output
        """
        fmt = u'{{:.{}g}}'.format(sig)
        return fmt.format(value)



from . import comparators


def get_default_equators():
    return (
        comparators.SequenceEquator(),
        comparators.BytesEquator(),
        comparators.StrEquator(),
        comparators.SetEquator(),
        comparators.OrderedDictEquator(),
        comparators.MappingEquator(),
        comparators.ComplexEquator(),
        comparators.RealEquator(),
        comparators.IntegerEquator(),
        comparators.BoolEquator(),
        comparators.NoneEquator(),
        comparators.TupleEquator(),
    )



from abc import ABCMeta, abstractmethod
from typing import Optional, MutableMapping, Any

from pytray import tree

from . import archives
from . import exceptions
from . import records

__all__ = 'Saver', 'Loader', 'SnapshotLoader', 'LiveDepositor'


class Base(metaclass=ABCMeta):
    """Common base for loader and saver"""

    def __init__(self, historian):
        self._historian = historian

    def get_archive(self) -> archives.Archive:
        return self._historian.get_archive()

    def get_historian(self):
        return self._historian


class Saver(Base, metaclass=ABCMeta):

    @abstractmethod
    def ref(self, obj) -> records.Ref:
        """Get a persistent reference for the given object"""

    def encode(self, obj, types_schema=None, path=()):
        """Encode a type for archiving"""
        if types_schema is None:
            types_schema = []

        historian = self.get_historian()

        if historian.is_primitive(obj):
            # Deal with the special containers by encoding their values if need be
            return tree.transform(self.encode, obj, path, types_schema=types_schema)

        # Store by value
        helper = historian.get_helper(type(obj), auto_register=True)
        save_state = helper.save_instance_state(obj, self)
        assert historian.is_primitive(save_state), "Saved state must be one of the primitive types"
        types_schema.append((path, helper.TYPE_ID))
        return self.encode(save_state, types_schema, path)

    def save_state(self, obj):
        """Save the state of an object and return the encoded state ready to be archived in a
        record"""
        state_types = []
        saved_state = self.encode(obj, state_types)
        return {records.STATE: saved_state, records.STATE_TYPES: state_types}


class Loader(Base, metaclass=ABCMeta):

    def decode(self, encoded, type_schema: dict = None, path=(), created_callback=None):
        try:
            type_id = type_schema[path]
        except KeyError:
            return self._unpack(encoded, type_schema, path, created_callback)
        else:
            saved_state = encoded
            helper = self.get_historian().get_helper(type_id)
            if helper.IMMUTABLE:
                saved_state = self._unpack(encoded, type_schema, path, created_callback)

            new_obj = helper.new(saved_state)
            assert new_obj is not None, \
                "Helper '{}' failed to create a class given state '{}'".format(
                    helper.__class__, saved_state)
            if created_callback is not None:
                created_callback(path, new_obj)

            if not helper.IMMUTABLE:
                saved_state = self._unpack(encoded, type_schema, path, created_callback)

            helper.load_instance_state(new_obj, saved_state, self)
            return new_obj

    def _unpack(self,
                encoded_saved_state,
                type_schema: dict = None,
                path=(),
                created_callback=None):
        """Unpack a saved state expanding any contained objects"""
        return tree.transform(self.decode,
                              encoded_saved_state,
                              path,
                              type_schema=type_schema,
                              created_callback=created_callback)


class LiveDepositor(Saver, Loader):
    """Depositor with strategy that all objects that get referenced should be saved"""

    def ref(self, obj) -> Optional[records.Ref]:
        if obj is None:
            return None

        try:
            # Try getting it from the transaction as there may be one from an in-progress save.
            # We can't use historian.get_ref here because we _only_ want one that's currently
            # being saved or we should try saving it as below to ensure it's up to date
            return self._historian.current_transaction().get_reference_for_live_object(obj)
        except exceptions.NotFound:
            # Then we have to save it and get the resulting reference
            return self._historian._save_object(obj, self).get_reference()

    def load(self, reference: records.Ref):
        try:
            return self._historian.get_obj(reference.obj_id)
        except exceptions.NotFound:
            return self._historian._load_object(reference.obj_id, self)

    def load_from_record(self, record):
        with self._historian.transaction() as trans:

            def created(path, new_obj):
                """Called each time an object is created whilst decoding"""
                # For the root object, put it into the transaction as a live object
                if not path:
                    trans.insert_live_object(new_obj, record)

            norm_schema = {tuple(path): type_id for path, type_id in record.state_types}
            return self.decode(record.state, norm_schema, created_callback=created)

    def update_from_record(self, obj, record) -> bool:
        historian = self.get_historian()
        helper = historian.get_helper(type(obj))
        with historian.transaction() as trans:
            # Make sure the record is in the transaction with the object
            trans.insert_live_object(obj, record)

            norm_schema = {tuple(path): type_id for path, type_id in record.state_types}
            saved_state = self._unpack(record.state, norm_schema)
            helper.load_instance_state(obj, saved_state, self)
            return True

    def save_from_builder(self, obj, builder):
        """Save a live object"""
        assert builder.snapshot_hash is not None, \
            "The snapshot hash must be set on the builder before saving"
        historian = self.get_historian()

        with historian.transaction() as trans:
            # Insert the object into the transaction so others can refer to it
            ref = records.Ref(builder.obj_id, builder.version)
            trans.insert_live_object_reference(ref, obj)

            # Deal with a possible object creator
            if builder.version == 0:
                try:
                    creator = historian.get_creator(obj)
                except exceptions.NotFound:
                    pass
                else:
                    # Found one
                    builder.extras[records.ExtraKeys.CREATED_BY] = self.ref(creator).obj_id

            # Now ask the object to save itself and create the record
            builder.update(self.save_state(obj))
            record = builder.build()

            # Insert the record into the transaction
            trans.insert_live_object(obj, record)
            trans.stage(record)

        return record


class SnapshotLoader(Loader):
    """Responsible for loading snapshots.  This object should not be reused and only
    one external call to `load` should be made.  This is because it keeps an internal
    cache."""

    def __init__(self, historian):
        super().__init__(historian)
        self._snapshots = {}  # type: MutableMapping[archives.Ref, Any]

    def load(self, ref: records.Ref):
        if not isinstance(ref, records.Ref):
            raise TypeError(ref)

        try:
            return self._snapshots[ref]
        except KeyError:
            record = self.get_archive().load(ref)
            if record.is_deleted_record():
                snapshot = None
            else:
                snapshot = self.load_from_record(record)

            # Cache it
            self._snapshots[ref] = snapshot
            return snapshot

    def load_from_record(self, record: records.DataRecord):
        with self._historian.transaction() as trans:
            norm_schema = {tuple(path): type_id for path, type_id in record.state_types}
            obj = self.decode(record.state, norm_schema)
            trans.insert_snapshot(obj, record.get_reference())
            return obj



import tempfile
import typing
from typing import Optional, Sequence
import uuid

from bidict import bidict
import bson
import gridfs
import pymongo
import pymongo.database
import pymongo.errors

from . import archives
from . import builtins
from . import depositors
from . import exceptions
from . import helpers
from . import records

__all__ = 'MongoArchive', 'GridFsFile'

OBJ_ID = records.OBJ_ID
TYPE_ID = records.TYPE_ID
CREATION_TIME = 'ctime'
VERSION = 'ver'
STATE = 'state'
STATE_TYPES = 'state_type'
SNAPSHOT_HASH = 'hash'
SNAPSHOT_TIME = 'stime'
EXTRAS = records.EXTRAS


class ObjectIdHelper(helpers.TypeHelper):
    TYPE = bson.ObjectId
    TYPE_ID = uuid.UUID('bdde0765-36d2-4f06-bb8b-536a429f32ab')

    def yield_hashables(self, obj, hasher):
        yield obj.binary

    def eq(self, one, other) -> bool:  # pylint: disable=invalid-name
        return one.__eq__(other)

    def save_instance_state(self, obj, _depositor):
        return obj

    def load_instance_state(self, obj, saved_state, _depositor):
        return obj.__init__(saved_state)


class MongoArchive(archives.BaseArchive[bson.ObjectId]):
    ID_TYPE = bson.ObjectId

    DATA_COLLECTION = 'data'
    META_COLLECTION = 'meta'

    # Here we map the data record property names onto ones in our entry format.
    # If a record property doesn't appear here it means the name says the same
    KEY_MAP = bidict({
        records.OBJ_ID: OBJ_ID,
        records.TYPE_ID: TYPE_ID,
        records.CREATION_TIME: CREATION_TIME,
        records.VERSION: VERSION,
        records.STATE: STATE,
        records.STATE_TYPES: STATE_TYPES,
        records.SNAPSHOT_HASH: SNAPSHOT_HASH,
        records.SNAPSHOT_TIME: SNAPSHOT_TIME,
        records.EXTRAS: EXTRAS,
    })

    META = 'meta'

    @classmethod
    def get_types(cls) -> Sequence:
        return ObjectIdHelper(), GridFsFile

    def __init__(self, database: pymongo.database.Database):
        self._data_collection = database[self.DATA_COLLECTION]
        self._meta_collection = database[self.META_COLLECTION]
        self._file_bucket = gridfs.GridFSBucket(database)
        self._create_indices()

    def _create_indices(self):
        # Make sure that no two entries can share the same object id and version
        self._data_collection.create_index([(self.KEY_MAP[records.OBJ_ID], pymongo.ASCENDING),
                                            (self.KEY_MAP[records.VERSION], pymongo.ASCENDING)],
                                           unique=True)

    def create_archive_id(self):  # pylint: disable=no-self-use
        return bson.ObjectId()

    def construct_archive_id(self, value) -> bson.ObjectId:
        if not isinstance(value, str):
            raise TypeError("Cannot construct an ObjectID from a '{}'".format(type(value)))
        return bson.ObjectId(value)

    def create_file(self, filename: str = None, encoding: str = None):
        return GridFsFile(self._file_bucket, filename, encoding)

    def get_gridfs_bucket(self) -> gridfs.GridFSBucket:
        return self._file_bucket

    def save(self, record: records.DataRecord):
        self.save_many([record])
        return record

    def save_many(self, records: typing.List[records.DataRecord]):
        # Generate the entries for our collection collecting the metadata that we gathered
        entries = [self._to_entry(record) for record in records]
        try:
            self._data_collection.insert_many(entries)
        except pymongo.errors.BulkWriteError as exc:
            write_errors = exc.details['writeErrors']
            if write_errors:
                raise exceptions.ModificationError(
                    "You're trying to rewrite history, that's not allowed!")
            raise  # Otherwise just raise what we got

    def load(self, reference: records.Ref) -> records.DataRecord:
        if not isinstance(reference, records.Ref):
            raise TypeError(reference)

        results = list(
            self._data_collection.find({
                self.KEY_MAP[records.OBJ_ID]: reference.obj_id,
                self.KEY_MAP[records.VERSION]: reference.version,
            }))
        if not results:
            raise exceptions.NotFound("Snapshot id '{}' not found".format(reference))
        return self._to_record(results[0])

    def get_snapshot_refs(self, obj_id):
        results = self._data_collection.find({OBJ_ID: obj_id}, sort=[(VERSION, pymongo.ASCENDING)])
        if not results:
            return []

        return [records.Ref(obj_id, result[VERSION]) for result in results]

    def get_meta(self, obj_id):
        assert isinstance(obj_id, bson.ObjectId), "Must pass an ObjectId"
        return self._meta_collection.find_one({'_id': obj_id}, projection={'_id': False})

    def set_meta(self, obj_id, meta):
        found = self._meta_collection.replace_one({'_id': obj_id}, meta, upsert=True)
        if not found:
            raise exceptions.NotFound("No record with snapshot id '{}' found".format(obj_id))

    def update_meta(self, obj_id, meta):
        self._meta_collection.update_one({'_id': obj_id}, {'$set': meta}, upsert=True)

    # pylint: disable=too-many-arguments
    def find(self,
             obj_id: Optional[bson.ObjectId] = None,
             type_id=None,
             _created_by=None,
             _copied_from=None,
             version=-1,
             state=None,
             snapshot_hash=None,
             meta=None,
             limit=0,
             sort=None,
             skip=0):
        pipeline = self._get_pipeline(obj_id=obj_id,
                                      type_id=type_id,
                                      _created_by=_created_by,
                                      _copied_from=_copied_from,
                                      version=version,
                                      state=state,
                                      snapshot_hash=snapshot_hash,
                                      meta=meta)

        if skip:
            pipeline.append({'$skip': skip})

        if limit:
            pipeline.append({'$limit': limit})

        if sort:
            if not isinstance(sort, dict):
                sort_dict = {sort: 1}
            else:
                sort_dict = sort
            sort_dict = self._remap(sort_dict)
            pipeline.append({'$sort': sort_dict})

        results = self._data_collection.aggregate(pipeline)

        for result in results:
            yield self._to_record(result)

    def count(self,
              obj_id: Optional[bson.ObjectId] = None,
              type_id=None,
              _created_by=None,
              _copied_from=None,
              version=-1,
              state=None,
              snapshot_hash=None,
              meta=None,
              limit=0):
        mfilter = {}
        if obj_id is not None:
            mfilter['obj_id'] = obj_id
        if type_id is not None:
            mfilter['type_id'] = type_id
        if state is not None:
            # If we are given a dict then expand as nested search criteria, e.g. {'state.colour': 'red'}
            if isinstance(state, dict):
                mfilter.update({"{}.{}".format(STATE, key): item for key, item in state.items()})
            else:
                mfilter[STATE] = state
        if snapshot_hash is not None:
            mfilter[self.KEY_MAP[records.SNAPSHOT_HASH]] = snapshot_hash
        if version == -1:
            # For counting we don't care which version we get we just want there to be only 1
            # counted per obj_id so select the first
            mfilter[VERSION] = 0
        else:
            mfilter[VERSION] = version

        pipeline = [{'$match': mfilter}]

        if meta:
            pipeline.append({
                '$lookup': {
                    'from': self._meta_collection.name,
                    'localField': OBJ_ID,
                    'foreignField': '_id',
                    'as': '_meta'
                }
            })
            # _meta should only contain at most one entry per document i.e. the metadata for
            # that object.  So check that for the search criteria
            pipeline.append(
                {'$match': {'_meta.0.{}'.format(key): value for key, value in meta.items()}})

        if limit:
            pipeline.append({'$limit': limit})

        pipeline.append({'$count': "total"})
        result = next(self._data_collection.aggregate(pipeline))

        return result['total']

    def _get_pipeline(self,
                      obj_id: Optional[bson.ObjectId] = None,
                      type_id=None,
                      _created_by=None,
                      _copied_from=None,
                      version=-1,
                      state=None,
                      snapshot_hash=None,
                      meta=None):
        """Get a pipeline that would perform the given search.  Can be used directly in an aggregate call"""
        mfilter = {}
        if obj_id is not None:
            mfilter['obj_id'] = obj_id
        if type_id is not None:
            mfilter['type_id'] = type_id
        if state is not None:
            # If we are given a dict then expand as nested search criteria, e.g. {'state.colour': 'red'}
            if isinstance(state, dict):
                predicates = []
                for key, value in state.items():
                    if key.startswith('$'):
                        predicates.append({key: value})
                    else:
                        mfilter.update({"{}.{}".format(STATE, key): value})
                if predicates:
                    mfilter.update({'$and': [{'state': predicate} for predicate in predicates]})
            else:
                mfilter[STATE] = state
        if snapshot_hash is not None:
            mfilter[self.KEY_MAP[records.SNAPSHOT_HASH]] = snapshot_hash
        if version is not None and version != -1:
            mfilter[VERSION] = version

        pipeline = [{'$match': mfilter}]

        if version == -1:
            # Join with a collection that is grouped to get the maximum version for each object ID
            # then only take the the matching documents
            pipeline.append({
                '$lookup': {
                    'from': self.DATA_COLLECTION,
                    'let': {
                        'obj_id': '$obj_id',
                        'ver': '$ver'
                    },
                    'pipeline': [
                        # Get the maximum version
                        {
                            '$group': {
                                '_id': '$obj_id',
                                'ver': {
                                    '$max': '$ver'
                                }
                            }
                        },
                        # Then match these with the obj id and version in our collection
                        {
                            '$match': {
                                '$expr': {
                                    '$and': [
                                        {
                                            '$eq': ['$_id', '$$obj_id']
                                        },
                                        {
                                            '$eq': ['$ver', '$$ver']
                                        },
                                    ]
                                }
                            }
                        }
                    ],
                    'as': "max_version"
                }
            })
            # Finally select those from our collection that have a 'max_version' array entry
            pipeline.append({"$match": {"max_version": {"$ne": []}}},)

        if meta:
            pipeline.append({
                '$lookup': {
                    'from': self._meta_collection.name,
                    'localField': OBJ_ID,
                    'foreignField': '_id',
                    'as': '_meta'
                }
            })
            # _meta should only contain at most one entry per document i.e. the metadata for
            # that object.  So check that for the search criteria
            pipeline.append(
                {'$match': {'_meta.0.{}'.format(key): value for key, value in meta.items()}})

        return pipeline

    def _to_record(self, entry) -> records.DataRecord:
        """Convert a MongoDB data collection entry to a DataRecord"""
        record_dict = records.DataRecord.defaults()

        # Invert our mapping of keys back to the data record property names and update over any defaults
        record_dict.update({
            recordkey: entry[dbkey] for recordkey, dbkey in self.KEY_MAP.items() if dbkey in entry
        })
        decoded_state = self._decode_state(record_dict[records.STATE])
        record_dict[records.STATE] = decoded_state

        return records.DataRecord(**record_dict)

    def _to_entry(self, record: records.DataRecord) -> dict:
        """Convert a DataRecord to a MongoDB data collection entry"""
        defaults = records.DataRecord.defaults()
        entry = {}
        for key, item in record._asdict().items():
            if key == records.STATE:
                entry[self.KEY_MAP[key]] = self._encode_state(item)
            else:
                # Exclude entries that have the default value
                if not (key in defaults and defaults[key] == item):
                    entry[self.KEY_MAP[key]] = item

        return entry

    def _encode_state(self, entry):
        if isinstance(entry, list):
            return [self._encode_state(item) for item in entry]

        if isinstance(entry, dict):
            return {key: self._encode_state(item) for key, item in entry.items()}

        return entry

    def _decode_state(self, entry):
        if isinstance(entry, dict):
            return {key: self._decode_state(item) for key, item in entry.items()}

        if isinstance(entry, list):
            return [self._decode_state(item) for item in entry]

        return entry

    def _remap(self, record_dict: dict) -> dict:
        """Given a dictionary return a new dictionary with the key names that we use"""
        remapped = {}
        for key, value in record_dict.items():
            split_key = key.split('.')
            split_key[0] = self.KEY_MAP[split_key[0]]
            remapped['.'.join(split_key)] = value
        return remapped


class GridFsFile(builtins.BaseFile):
    TYPE_ID = uuid.UUID('3bf3c24e-f6c8-4f70-956f-bdecd7aed091')
    ATTRS = '_persistent_id', '_file_id'

    def __init__(self,
                 file_bucket: gridfs.GridFSBucket,
                 filename: str = None,
                 encoding: str = None):
        super().__init__(filename, encoding)
        self._file_store = file_bucket
        self._file_id = None
        self._persistent_id = bson.ObjectId()
        self._buffer_file = self._create_buffer_file()

    def open(self, mode='r', **kwargs):
        self._ensure_buffer()
        if 'b' not in mode:
            kwargs.setdefault('encoding', self.encoding)
        return open(self._buffer_file, mode, **kwargs)

    def save_instance_state(self, saver: depositors.Saver):
        filename = self.filename or ""
        with open(self._buffer_file, 'rb') as fstream:
            self._file_id = self._file_store.upload_from_stream(filename, fstream)

        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader: depositors.Loader):
        super().load_instance_state(saved_state, loader)
        self._file_store = loader.get_archive().get_gridfs_bucket()  # type: gridfs.GridFSBucket
        # Don't copy the file over now, do it lazily when the file is first opened
        self._buffer_file = None

    def _ensure_buffer(self):
        if self._buffer_file is None:
            if self._file_id is not None:
                self._update_buffer()
            else:
                self._create_buffer_file()

    def _update_buffer(self):
        self._buffer_file = self._create_buffer_file()
        with open(self._buffer_file, 'wb') as fstream:
            self._file_store.download_to_stream(self._file_id, fstream)

    def _create_buffer_file(self):
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        tmp_path = tmp_file.name
        tmp_file.close()
        return tmp_path



from argparse import Namespace
import pathlib
import uuid

import mincepy
import mincepy.builtins

from mincepy.testing import Car, Person


def test_path_helper(historian: mincepy.Historian):
    historian.register_type(mincepy.common_helpers.PathHelper())

    class File(mincepy.BaseSavableObject):
        TYPE_ID = uuid.UUID('8d645bb8-4657-455b-8b61-8613bc8a0acf')
        ATTRS = ('path',)

        def __init__(self, path):
            super(File, self).__init__()
            self.path = path

    file = File(pathlib.Path('some_path'))
    file_id = historian.save(file)
    del file
    loaded = historian.load(file_id)
    assert loaded.path == pathlib.Path('some_path')


def test_tuple_helper(historian: mincepy.Historian):
    historian.register_type(mincepy.common_helpers.TupleHelper())

    container = mincepy.builtins.List()
    container.append((Car('ferrari'),))

    container_id = historian.save(container)
    del container
    loaded = historian.load(container_id)
    assert loaded[0][0] == Car('ferrari')


def test_namespace_helper(historian: mincepy.Historian):
    historian.register_type(mincepy.common_helpers.NamespaceHelper())

    car = Car('fiat', '500')
    dinner = Namespace(food='pizza', drink='wine', cost=10.94, car=car, host=Person('Martin', 34))
    dinner_id = historian.save(dinner)
    del dinner

    loaded = historian.load(dinner_id)
    assert loaded.car == car
    assert loaded.host == Person('Martin', 34)
    assert loaded.food == 'pizza'
    assert loaded.cost == 10.94

    loaded.guest = Person('Sonia', 30)
    historian.save(loaded)
    del loaded

    reloaded = historian.load(dinner_id)
    assert reloaded.guest == Person('Sonia', 30)



from .archives import *
from .builtins import *
from .base_savable import *
from .archive_factory import *
from .convenience import *
from .comparators import *
from .depositors import *
from .exceptions import *
from .helpers import *
from .historians import *
from .history import *
from .process import *
from .records import *
from .refs import *
from .types import *
from .version import *
from . import analysis
from . import common_helpers
from . import mongo
from . import testing
from . import utils

_ADDITIONAL = ('analysis', 'mongo', 'buitins', 'common_helpers', 'testing', 'utils')

__all__ = (archives.__all__ + comparators.__all__ + depositors.__all__ + exceptions.__all__ +
           historians.__all__ + convenience.__all__ + process.__all__ + types.__all__ +
           helpers.__all__ + version.__all__ + history.__all__ + archive_factory.__all__ +
           refs.__all__ + records.__all__ + base_savable.__all__ + builtins.__all__ + _ADDITIONAL)



from abc import ABCMeta, abstractmethod
import typing
from typing import Sequence

from .records import DataRecord, Ref

__all__ = 'Archive', 'ASCENDING', 'DESCENDING'

IdT = typing.TypeVar('IdT')  # The archive ID type

# Sort options
ASCENDING = 1
DESCENDING = -1


class Archive(typing.Generic[IdT], metaclass=ABCMeta):

    @classmethod
    def get_types(cls) -> Sequence:
        """This method allows the archive to return either types or type helper that the historian
        should support.  A common example is the type helper for the object id type"""
        return tuple()

    @classmethod
    def get_extra_primitives(cls) -> tuple:
        """Can optionally return a list of types that are treated as primitives i.e. considered to be
        storable and retrievable directly without encoding."""
        return tuple()

    @classmethod
    @abstractmethod
    def get_id_type(cls) -> typing.Type[IdT]:
        """Get the type used as an ID by this archive"""

    @abstractmethod
    def create_archive_id(self) -> IdT:
        """Create a new archive id"""

    @abstractmethod
    def construct_archive_id(self, value) -> IdT:
        """If it's possible, construct an archive value from the passed value.
        This is useful as a convenience to the user if, say, the archive id can be
        constructed from a string.  Raise TypeError or ValueError if this is not possible
        for the given value.
        """

    @abstractmethod
    def create_file(self, filename: str = None, encoding: str = None):
        """Create a new file object specific for this archive type"""

    @abstractmethod
    def save(self, record: DataRecord):
        """Save a data record to the archive"""

    @abstractmethod
    def save_many(self, records: typing.Sequence[DataRecord]):
        """Save many data records to the archive"""

    # region Metadata

    @abstractmethod
    def get_meta(self, obj_id: IdT):
        """Get the metadata for the given object snapshot."""

    @abstractmethod
    def set_meta(self, obj_id: IdT, meta):
        """Set the metadata on on the object with the corresponding id"""

    @abstractmethod
    def update_meta(self, obj_id: IdT, meta):
        """Update the metadata on the object with the corresponding id"""

    # endregion

    @abstractmethod
    def load(self, reference: Ref[IdT]) -> DataRecord:
        """Load a snapshot of an object with the given reference"""

    @abstractmethod
    def history(self, obj_id: IdT, idx_or_slice) -> [DataRecord, typing.List[DataRecord]]:
        """Load the snapshot records for a particular object, can return a single or multiple records"""

    @abstractmethod
    def get_snapshot_refs(self, obj_id: IdT) -> typing.Sequence[Ref[IdT]]:
        """Returns a list of time ordered snapshot references"""

    # pylint: disable=too-many-arguments
    @abstractmethod
    def find(self,
             obj_id=None,
             type_id=None,
             created_by=None,
             copied_from=None,
             version=-1,
             state=None,
             snapshot_hash=None,
             meta=None,
             limit=0,
             sort=None,
             skip=0):
        """Find records matching the given criteria"""

    @abstractmethod
    def count(self,
              obj_id=None,
              type_id=None,
              created_by=None,
              copied_from=None,
              version=-1,
              state=None,
              snapshot_hash=None,
              meta=None,
              limit=0):
        """Count the number of entries that match the given query"""


class BaseArchive(Archive[IdT]):
    ID_TYPE = None  # type: typing.Type[IdT]

    @classmethod
    def get_id_type(cls) -> typing.Type[IdT]:
        assert cls.ID_TYPE, "The ID type has not been set on this archive"
        return cls.ID_TYPE

    def save_many(self, records: typing.Sequence[DataRecord]):
        """
        This will save records one by one but subclass may want to override this behaviour if
        they can save multiple records at once.
        """
        for record in records:
            self.save(record)

    def history(self, obj_id: IdT, idx_or_slice) -> [DataRecord, typing.List[DataRecord]]:
        refs = self.get_snapshot_refs(obj_id)[idx_or_slice]
        if len(refs) > 1:
            return [self.load(ref) for ref in refs]

        # Single one
        return self.load(refs[0])

    def construct_archive_id(self, value) -> IdT:
        raise TypeError("Not possible to construct an archive id from '{}'".format(type(value)))



import mincepy

from mincepy.testing import Car


def test_transaction_snapshots(historian: mincepy.Historian):
    ferrari = Car('ferrari')
    ferrari_id = historian.save(ferrari, return_sref=True)

    with historian.transaction():
        ferrari_snapshot_1 = historian.load_snapshot(ferrari_id)
        with historian.transaction():
            ferrari_snapshot_2 = historian.load_snapshot(ferrari_id)
            # Reference wise they should be unequal
            assert ferrari_snapshot_1 is not ferrari_snapshot_2
            assert ferrari is not ferrari_snapshot_1
            assert ferrari is not ferrari_snapshot_2

            # Value wise they should be equal
            assert ferrari == ferrari_snapshot_1
            assert ferrari == ferrari_snapshot_2

        # Now check within the same transaction the result is the same
        ferrari_snapshot_2 = historian.load_snapshot(ferrari_id)
        # Reference wise they should be unequal
        assert ferrari_snapshot_1 is not ferrari_snapshot_2
        assert ferrari is not ferrari_snapshot_1
        assert ferrari is not ferrari_snapshot_2

        # Value wise they should be equal
        assert ferrari == ferrari_snapshot_1
        assert ferrari == ferrari_snapshot_2


def test_transaction_records(historian: mincepy.Historian):
    """Make sure that records within a transaction are not recreated at each save"""
    with historian.transaction():
        ferrari = Car('ferrari')

        # Save and get the record for the ferrari
        ferrari_id = historian.save(ferrari)
        record = historian.get_current_record(ferrari)

        ferrari_id2 = historian.save(ferrari)
        assert ferrari_id == ferrari_id2
        assert historian.get_current_record(ferrari) is record

        loaded = historian.load(ferrari_id)
        assert loaded is ferrari


def test_find(historian: mincepy.Historian):
    honda_id = historian.save(Car('honda'))
    zonda_id = historian.save(Car('zonda'))
    porsche_id = historian.save(Car('porsche'))

    cars = list(historian.find(Car))
    assert len(cars) == 3

    makes = [car.make for car in cars]
    assert 'honda' in makes
    assert 'zonda' in makes
    assert 'porsche' in makes


def test_update(historian: mincepy.Historian):
    car = Car('ferrari', 'red')
    historian.save(car)

    # Simulate saving the car from another connection
    honda = Car('honda', 'black')
    historian.save(honda)
    honda_record = historian.get_current_record(honda)

    archive = historian.get_archive()
    record = historian.get_current_record(car)
    builder = record.child_builder(obj_id=historian.get_obj_id(car),
                                   snapshot_hash=honda_record.snapshot_hash,
                                   state=honda_record.state,
                                   state_types=honda_record.state_types)
    archive.save(builder.build())

    # Now update and check the state
    historian.sync(car)
    assert car.make == 'honda'
    assert car.colour == 'black'

    # Also, check the cached record
    car_record = historian.get_current_record(car)
    assert car_record.snapshot_hash == honda_record.snapshot_hash
    assert car_record.state == honda_record.state



from collections import namedtuple
import contextlib
import copy
import getpass
import logging
import socket
import typing
from typing import MutableMapping, Any, Optional, Mapping
import weakref

from . import archives
from . import builtins
from . import defaults
from . import depositors
from . import refs
from . import exceptions
from . import helpers
from . import process
from . import records
from . import types
from . import type_registry
from . import utils
from .transactions import RollbackTransaction, Transaction, LiveObjects

__all__ = 'Historian', 'ObjectEntry'

logger = logging.getLogger(__name__)
ObjectEntry = namedtuple('ObjectEntry', 'ref obj')


class Historian:

    def __init__(self, archive: archives.Archive, equators=()):
        self._archive = archive
        self._equator = types.Equator(defaults.get_default_equators() + equators)
        # Register default types
        self._type_registry = type_registry.TypeRegistry()
        self.register_type(refs.ObjRef)
        self.register_type(helpers.SnapshotRefHelper())
        self.register_types(archive.get_types())

        # Snapshot objects -> reference. Objects that were loaded from historical snapshots
        self._snapshots_objects = utils.WeakObjectIdDict()  # type: MutableMapping[Any, records.Ref]
        self._live_objects = LiveObjects()

        # Staged objects that have been created but not saved
        self._creators = utils.WeakObjectIdDict()  # type: MutableMapping[typing.Any, Any]
        self._type_ids = {}

        self._transactions = None

        self._saving_set = set()

        self._user = getpass.getuser()
        self._hostname = socket.gethostname()

    def get_archive(self):
        return self._archive

    def created(self, obj):
        """Called when an object is created.  The historian tracks the creator for saving when
        the object is saved"""
        creator = process.Process.current_process()
        if creator is not None:
            self._creators[obj] = creator

    def create_file(self, filename: str = None, encoding: str = None) -> builtins.BaseFile:
        """Create a new file.  The historian will supply file type compatible with the archive in use."""
        return self._archive.create_file(filename, encoding)

    def save(self, *objs, with_meta=None, return_sref=False):
        """Save or more objects in the history producing corresponding object identifiers

        :param objs: the object(s) to save
        :param with_meta: the object(s) metadata (has to be a sequence if more than one)
        :param return_sref: if True will return a snapshot reference, otherwise just the object id
        """
        if with_meta is not None:
            if len(objs) == 1:
                with_meta = (with_meta,)
            else:
                assert len(objs) == len(with_meta), \
                    "The metadata should be a sequence with the same number of entries as the number of objects"
        else:
            with_meta = [None] * len(objs)

        with self.transaction():
            ids = []
            for obj, meta in zip(objs, with_meta):
                ids.append(self.save_one(obj, meta, return_sref))
            if len(ids) == 1:
                return ids[0]

            return ids

    def save_one(self, obj, with_meta=None, return_sref=False):
        """Save the object in the history producing a unique id"""
        if obj in self._snapshots_objects:
            raise exceptions.ModificationError(
                "Cannot save a snapshot object, that would rewrite history!")

        record = self.save_object(obj)
        if with_meta is not None:
            self._archive.set_meta(record.obj_id, with_meta)

        if return_sref:
            return record.get_reference()

        return record.obj_id

    def replace(self, old, new):
        """Replace a live object with a new version.

        This is especially useful if you have made a copy of an object and modified it but you want to
        continue the history of the object as the original rather than a brand new object.  Then just
        replace the old object with the new one by calling this function.
        """
        assert not self.current_transaction(
        ), "Can't replace during a transaction for the time being"
        assert isinstance(new, type(old)), "Can't replace type '{} with type '{}!".format(
            type(old), type(new))

        # Get the current record and replace the object with the new one
        record = self._live_objects.get_record(old)
        self._live_objects.delete(old)
        self._live_objects.insert(new, record)

        # Make sure creators is correct as well
        try:
            self._creators[new] = self._creators.pop(old)
        except KeyError:
            pass

    def load_snapshot(self, reference: records.Ref) -> Any:
        return depositors.SnapshotLoader(self).load(reference)

    def load(self, *obj_ids_or_refs):
        """Load object(s) or snapshot(s)."""
        loaded = []
        for entry in obj_ids_or_refs:
            loaded.append(self.load_one(entry))

        if len(obj_ids_or_refs) == 1:
            return loaded[0]

        return loaded

    def load_one(self, obj_id_or_ref):
        """Load one object or shot from the database"""
        if isinstance(obj_id_or_ref, records.Ref):
            return self.load_snapshot(obj_id_or_ref)

        return self.load_object(obj_id_or_ref)

    def sync(self, obj) -> bool:
        """Update an object with the latest state in the database.
        If there is no new version in the archive then the current version remains
        unchanged including any modifications.

        :return: True if the object was updated, False otherwise
        """
        try:
            obj_id = self.get_obj_id(obj)
        except exceptions.NotFound:
            # Not found, so the object is as up to date as can be i.e. never saved!
            return False

        ref = self._get_latest_snapshot_reference(obj_id)
        archive_record = self._archive.load(ref)
        if archive_record.is_deleted_record():
            raise exceptions.ObjectDeleted("Object with id '{}' has been deleted".format(obj_id))

        if ref.version == self.get_snapshot_ref(obj).version:
            # Nothing has changed
            return False

        # The one in the archive is newer, so use that
        depositor = depositors.LiveDepositor(self)
        return depositor.update_from_record(obj, archive_record)

    def copy(self, obj):
        """Create a shallow copy of the object, save that copy and return it"""
        with self.transaction() as trans:
            record = self._save_object(obj, depositors.LiveDepositor(self))
            copy_builder = record.copy_builder(obj_id=self._archive.create_archive_id())
            self._record_builder_created(copy_builder)

            # Copy the object and record
            obj_copy = copy.copy(obj)
            obj_copy_record = copy_builder.build()

            # Insert all the new objects into the transaction
            trans.insert_live_object(obj_copy, obj_copy_record)
            trans.stage(obj_copy_record)

        return obj_copy

    def delete(self, obj):
        """Delete a live object"""
        record = self.get_current_record(obj)
        with self.transaction() as trans:
            builder = records.make_deleted_builder(record)
            deleted_record = self._record_builder_created(builder).build()
            trans.stage(deleted_record)
        self._live_objects.delete(obj)

    def history(
            self,
            obj_or_obj_id,
            idx_or_slice='*',
            as_objects=True) -> [typing.Sequence[ObjectEntry], typing.Sequence[records.DataRecord]]:
        """
        Get a sequence of object ids and instances from the history of the given object.

        :param obj_or_obj_id: The instance or id of the object to get the history for
        :param idx_or_slice: The particular index or a slice of which historical versions to get
        :param as_objects: if True return the object instances, otherwise returns the DataRecords

        Example:
        >>> historian = get_historian()
        >>> car = Car('ferrari', 'white')
        >>> car_id = historian.save(car)
        >>> car.colour = 'red'
        >>> historian.save(car)
        >>> history = historian.history(car_id)
        >>> len(history)
        2
        >>> history[0].obj.colour == 'white'
        True
        >>> history[1].obj.colour == 'red'
        True
        >>> history[1].obj is car
        """
        obj_id = self._ensure_obj_id(obj_or_obj_id)
        snapshot_refs = self._archive.get_snapshot_refs(obj_id)
        indices = utils.to_slice(idx_or_slice)
        to_get = snapshot_refs[indices]
        if as_objects:
            return [ObjectEntry(ref, self.load_snapshot(ref)) for ref in to_get]

        return [self._archive.load(ref) for ref in to_get]

    def load_object(self, obj_id):
        return self._load_object(obj_id, depositors.LiveDepositor(self))

    def save_object(self, obj) -> records.DataRecord:
        return self._save_object(obj, depositors.LiveDepositor(self))

    # region Metadata

    def get_meta(self, obj_or_identifier) -> dict:
        """Get the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        """
        if isinstance(obj_or_identifier, self._archive.get_id_type()):
            obj_id = obj_or_identifier
        else:
            obj_id = self.get_obj_id(obj_or_identifier)

        return self._archive.get_meta(obj_id)

    def set_meta(self, obj_or_identifier, meta: Optional[Mapping]):
        """Set the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        :param meta: the metadata dictionary
        """
        if isinstance(obj_or_identifier, self._archive.get_id_type()):
            obj_id = obj_or_identifier
        else:
            obj_id = self.get_obj_id(obj_or_identifier)

        self._archive.set_meta(obj_id, meta)

    def update_meta(self, obj_or_identifier, meta: Mapping):
        """Update the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        :param meta: the metadata dictionary
        """
        obj_id = self._ensure_obj_id(obj_or_identifier)
        self._archive.update_meta(obj_id, meta)

    # endregion

    def get_current_record(self, obj) -> records.DataRecord:
        """Get a record for an object known to the historian"""
        trans = self.current_transaction()
        # Try the transaction first
        if trans:
            try:
                return trans.live_objects.get_record(obj)
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj)

    def get_obj_id(self, obj):
        """Get the object ID for a live object"""
        return self.get_current_record(obj).obj_id

    def get_obj(self, obj_id):
        """Get an object known to the historian"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_live_object(obj_id)
            except ValueError:
                pass

        return self._live_objects.get_object(obj_id)

    def get_snapshot_ref(self, obj):
        """Get the current snapshot reference for a live object"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_reference_for_live_object(obj)
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj).get_reference()

    def hash(self, obj):
        return self._equator.hash(obj)

    def eq(self, one, other):  # pylint: disable=invalid-name
        return self._equator.eq(one, other)

    # region Types

    @classmethod
    def is_trackable(cls, obj):
        """Determine if an object is trackable i.e. we can treat these as live objects and automatically
        keep track of their history when saving.  Ultimately this is determined by whether the type is
        weak referencable or not.
        """
        try:
            weakref.ref(obj)
            return True
        except TypeError:
            return False

    def is_primitive(self, obj):
        """Check if the object is one of the primitives and should be saved by value in the archive"""
        primitives = types.PRIMITIVE_TYPES + (
            self._archive.get_id_type(),) + self._archive.get_extra_primitives()
        return isinstance(obj, primitives)

    def is_obj_id(self, obj_id):
        return isinstance(obj_id, self._archive.get_id_type())

    def register_type(
        self, obj_class_or_helper: [helpers.TypeHelper, typing.Type[types.SavableObject]]
    ) -> helpers.TypeHelper:
        helper = self._type_registry.register_type(obj_class_or_helper)
        self._equator.add_equator(helper)
        return helper

    def register_types(self, obj_clases_or_helpers):
        for item in obj_clases_or_helpers:
            self.register_type(item)

    def get_obj_type_id(self, obj_type):
        return self._type_registry.get_type_id(obj_type)

    def get_obj_type(self, type_id):
        return self.get_helper(type_id).TYPE

    def get_helper(self, type_id_or_type, auto_register=False) -> helpers.TypeHelper:
        if auto_register and issubclass(type_id_or_type, types.SavableObject):
            self._ensure_compatible(type_id_or_type)

        return self._type_registry.get_helper(type_id_or_type)

    # endregion

    def find(self,
             obj_type=None,
             version: int = -1,
             state: dict = None,
             meta: dict = None,
             sort=None,
             limit=0,
             skip=0,
             as_objects=True):
        """Find entries in the archive

        :param obj_type: the object type to look for
        :param version: the version of the object to retrieve, -1 means latest
        :param state: the criteria on the state of the object to apply
        :param meta: the search criteria to apply on the metadata of the object
        :param sort: the sort criteria
        :param limit: the maximum number of results to return, 0 means unlimited
        :param skip: the page to get results from
        :param as_objects: if True returns the live object instances, False returns the DataRecords
        """
        type_id = obj_type
        if obj_type is not None:
            try:
                type_id = self.get_obj_type_id(obj_type)
            except TypeError:
                pass
        results = self._archive.find(type_id=type_id,
                                     state=state,
                                     version=version,
                                     meta=meta,
                                     sort=sort,
                                     limit=limit,
                                     skip=skip)
        if as_objects:
            for result in results:
                yield self.load(result.obj_id)
        else:
            yield from results

    def get_creator(self, obj_or_identifier):
        """Get the object that created the passed object"""
        if not self.is_obj_id(obj_or_identifier):
            # Object instance, try our creators cache
            try:
                return self._creators[obj_or_identifier]
            except KeyError:
                pass

        creator_id = self.created_by(obj_or_identifier)
        return self.load_object(creator_id)

    def created_by(self, obj_or_identifier):
        """Return the id of the object that created the passed object"""
        try:
            record = self.get_current_record(obj_or_identifier)
        except exceptions.NotFound:
            if not self.is_obj_id(obj_or_identifier):
                raise
            record = self._archive.load(self._get_latest_snapshot_reference(obj_or_identifier))

        return record.created_by

    def get_user_info(self) -> dict:
        user_info = {}
        if self._user:
            user_info[records.ExtraKeys.USER] = self._user
        if self._hostname:
            user_info[records.ExtraKeys.HOSTNAME] = self._hostname
        return user_info

    @contextlib.contextmanager
    def transaction(self):
        """Start a new transaction.  Will be nested if there is already one underway"""
        if self._transactions:
            # Start a nested one
            with self._transactions[-1].nested() as nested:
                self._transactions.append(nested)
                try:
                    yield nested
                finally:
                    popped = self._transactions.pop()
                    assert popped is nested
        else:
            # New transaction
            trans = Transaction()
            self._transactions = [trans]

            try:
                yield trans
            except RollbackTransaction:
                pass
            else:
                # Commit the transaction
                # Live objects
                self._live_objects.update(trans.live_objects)

                # Snapshots
                for ref, obj in trans.snapshots.items():
                    self._snapshots_objects[obj] = ref

                # Save any records that were staged for archiving
                if trans.staged:
                    self._archive.save_many(trans.staged)
            finally:
                assert len(self._transactions) == 1
                assert self._transactions[0] is trans
                self._transactions = None

    def current_transaction(self) -> Optional[Transaction]:
        """Get the current transaction if there is one, otherwise returns None"""
        if not self._transactions:
            return None
        return self._transactions[-1]

    def _get_latest_snapshot_reference(self, obj_id) -> records.Ref:
        """Given an object id this will return a reference to the latest snapshot"""
        try:
            return self._archive.get_snapshot_refs(obj_id)[-1]
        except IndexError:
            raise exceptions.NotFound("Object with id '{}' not found.".format(obj_id))

    def _load_object(self, obj_id, depositor: depositors.LiveDepositor):
        obj_id = self._ensure_obj_id(obj_id)

        with self.transaction() as trans:
            # Try getting the object from the our dict of up to date ones
            try:
                return trans.get_live_object(obj_id)
            except exceptions.NotFound:
                pass

            # Couldn't find it, so let's check if we have one and check if it is up to date
            ref = self._get_latest_snapshot_reference(obj_id)
            record = self._archive.load(ref)
            if record.is_deleted_record():
                raise exceptions.ObjectDeleted(
                    "Object with id '{}' has been deleted".format(obj_id))

            try:
                obj = self._live_objects.get_object(obj_id)
            except exceptions.NotFound:
                # Ok, just use the one from the archive
                return depositor.load_from_record(record)
            else:
                if record.version != self.get_snapshot_ref(obj).version:
                    # The one in the archive is newer, so use that
                    depositor.update_from_record(obj, record)

                return obj

    def _save_object(self, obj, depositor) -> records.DataRecord:
        with self.transaction() as trans:
            helper = self._ensure_compatible(type(obj))

            # Check if an object is already being saved in the transaction
            try:
                record = trans.get_record_for_live_object(obj)
                return record
            except exceptions.NotFound:
                pass

            with self._saving(obj):
                # Ok, have to save it
                current_hash = self.hash(obj)

                try:
                    # Let's see if we have a record at all
                    record = self._live_objects.get_record(obj)
                except exceptions.NotFound:
                    builder = self._create_builder(obj, dict(snapshot_hash=current_hash))
                    return depositor.save_from_builder(obj, builder)
                else:
                    if helper.IMMUTABLE:
                        logger.info("Tried to save immutable object with id '%s' again",
                                    record.obj_id)
                        return record

                    # Check if our record is up to date
                    with self.transaction() as transaction:
                        loaded_obj = depositors.SnapshotLoader(self).load_from_record(record)

                        if current_hash == record.snapshot_hash and self.eq(obj, loaded_obj):
                            # Objects identical
                            transaction.rollback()
                        else:
                            builder = record.child_builder(snapshot_hash=current_hash)
                            self._record_builder_created(builder)
                            record = depositor.save_from_builder(obj, builder)

                    return record

    def _create_builder(self, obj, additional=None):
        additional = additional or {}
        helper = self._ensure_compatible(type(obj))
        builder = records.DataRecord.new_builder(type_id=helper.TYPE_ID,
                                                 obj_id=self._archive.create_archive_id(),
                                                 version=0)
        self._record_builder_created(builder)
        builder.update(additional)
        return builder

    def _ensure_obj_id(self, obj_or_identifier):
        """
        This call will try and get an object id from the passed parameter.  There are three possibilities:
            1) It is passed an object ID in which case it will be returned directly
            2) It is passed an object instance, in which case it will try and get the id from memory
            3) It is passed a type that can be used as a constructor argument to the object id type in which
                case it will construct it and return the result
        """
        if self.is_obj_id(obj_or_identifier):
            return obj_or_identifier

        try:
            # Try creating it for the user by calling the constructor with the argument passed.
            # This helps for common obj id types which can be constructed from a string
            return self._archive.construct_archive_id(obj_or_identifier)
        except (ValueError, TypeError):
            # Maybe we've been passed an object
            return self.get_obj_id(obj_or_identifier)

    def _ensure_compatible(self, obj_type) -> helpers.TypeHelper:
        if obj_type not in self._type_registry:
            return self.register_type(obj_type)

        return self._type_registry.get_helper_from_obj_type(obj_type)

    @contextlib.contextmanager
    def _saving(self, obj):
        obj_id = id(obj)
        if obj_id in self._saving_set:
            raise RuntimeError(
                "The object is already being saved, this cannot be called twice and suggests "
                "a circular reference is being made")
        self._saving_set.add(obj_id)
        try:
            yield
        finally:
            self._saving_set.remove(obj_id)

    def _record_builder_created(self,
                                builder: records.DataRecordBuilder) -> records.DataRecordBuilder:
        """Update a data record builder with standard information."""
        builder.extras.update(self.get_user_info())
        return builder



__all__ = ('NotFound', 'ModificationError', 'ObjectDeleted')


class NotFound(Exception):
    """Raised when something can not be found in the history"""


class ModificationError(Exception):
    """Raised when a modification of the history encountered a problem"""


class ObjectDeleted(Exception):
    """Raise when the user tries to interact with a deleted object"""



"""Module for all the built in container and other types"""

from abc import ABCMeta, abstractmethod
import collections
from contextlib import contextmanager
from pathlib import Path
import shutil
from typing import BinaryIO, Optional
import uuid

from . import base_savable
from . import refs
from .utils import sync

__all__ = ('List', 'LiveList', 'LiveRefList', 'RefList', 'Str', 'Dict', 'RefDict', 'LiveDict',
           'LiveRefDict', 'BaseFile', 'File')


class _UserType(base_savable.BaseSavableObject, metaclass=ABCMeta):
    """Mixin for helping user types to be compatible with the historian.
    These typically have a .data member that stores the actual data (list, dict, str, etc)"""
    # pylint: disable=too-few-public-methods
    ATTRS = ('data',)
    data = None  # placeholder


class ObjProxy(_UserType):
    """A simple proxy for any object/data type which can also be a primitive"""
    TYPE_ID = uuid.UUID('d43c2db5-1e8c-428f-988f-8b198accde47')

    def __init__(self, data=None, historian=None):
        super(ObjProxy, self).__init__(historian)
        self.data = data

    def __call__(self):
        return self.data

    def assign(self, value):
        self.data = value


class Str(collections.UserString, _UserType):
    TYPE_ID = uuid.UUID('350f3634-4a6f-4d35-b229-71238ce9727d')

    def __init__(self, seq, historian=None):
        collections.UserString.__init__(self, seq)
        _UserType.__init__(self, historian)


class Reffer:
    """Mixin for types that want to be able to create references non-primitive types"""

    # pylint: disable=no-member, too-few-public-methods

    def _ref(self, obj):
        """Create a reference for a non-primitive, otherwise use the value"""
        if self._historian.is_primitive(obj):
            return obj

        return self._create_ref(obj)

    def _deref(self, obj):
        """Dereference a non-primitive, otherwise return the value"""
        if self._historian.is_primitive(obj):
            return obj

        return obj()

    def _create_ref(self, obj):
        """Create a reference for a given object"""
        return refs.ObjRef(obj, self._historian)


# region lists


class List(collections.UserList, _UserType):
    TYPE_ID = uuid.UUID('2b033f70-168f-4412-99ea-d1f131e3a25a')

    def __init__(self, initlist=None, historian=None):
        collections.UserList.__init__(self, initlist)
        _UserType.__init__(self, historian)


class RefList(collections.abc.MutableSequence, Reffer, _UserType):
    """A list that stores all entries as references in the database except primitives"""
    TYPE_ID = uuid.UUID('091efff5-136d-4ac2-bd59-28f50f151263')
    CONTAINER = list

    def __init__(self, init_list=None, historian=None):
        super().__init__(historian)
        self.data = []
        if init_list is not None:
            self.data = self.CONTAINER(self._ref(item) for item in init_list)

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __len__(self):
        return self.data.__len__()

    def insert(self, index, value):
        self.data.insert(index, self._ref(value))


class LiveList(collections.abc.MutableSequence, _UserType):
    """A list that is always in sync with the database"""
    TYPE_ID = uuid.UUID('c83e6206-cd29-4fda-bf76-11fce1681cd9')

    def __init__(self, init_list=None, historian=None):
        super(LiveList, self).__init__(historian)
        init_list = init_list or []
        self.data = RefList(init_list)

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        return proxy()

    @sync(save=True)
    def __setitem__(self, key, value):
        proxy = self.data[key]  # type: ObjProxy
        proxy.assign(value)

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]  # type: ObjProxy
        del self.data[key]
        self._historian.delete(proxy)

    @sync()
    def __len__(self):
        return len(self.data)

    @sync(save=True)
    def insert(self, index, value):
        proxy = self._create_proxy(value)
        self.data.insert(index, proxy)

    @sync(save=True)
    def append(self, value):
        proxy = self._create_proxy(value)
        self.data.append(proxy)

    def _create_proxy(self, obj):
        return ObjProxy(obj, self._historian)


class LiveRefList(RefList):
    """A live list that uses references to store objects"""
    TYPE_ID = uuid.UUID('98454806-c587-4fcc-a514-65fdefb0180d')
    CONTAINER = LiveList


# endregion

# region Dicts


class Dict(collections.UserDict, _UserType):
    TYPE_ID = uuid.UUID('a7584078-95b6-4e00-bb8a-b077852ca510')

    def __init__(self, *args, historian=None, **kwarg):
        collections.UserDict.__init__(self, *args, **kwarg)
        _UserType.__init__(self, historian)


class RefDict(collections.MutableMapping, Reffer, _UserType):
    """A dictionary that stores all values as references in the database."""
    TYPE_ID = uuid.UUID('c95f4c4e-766b-4dda-a43c-5fca4fd7bdd0')
    CONTAINER = dict

    def __init__(self, *args, **kwargs):
        super().__init__(kwargs.pop('historian', None))
        initial = dict(*args, **kwargs)
        if initial:
            self.data = self.CONTAINER({key: self._ref(value) for key, value in initial.items()})
        else:
            self.data = self.CONTAINER()

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()


class LiveDict(collections.MutableMapping, _UserType):
    TYPE_ID = uuid.UUID('740cc832-721c-4f85-9628-706257eb55b9')

    def __init__(self, *args, **kwargs):
        historian = kwargs.pop('historian', None)
        super().__init__(historian=historian)

        initial = dict(*args, **kwargs)
        self.data = RefDict({key: self._create_proxy(value) for key, value in initial.items()})

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        return proxy()

    @sync()
    def __setitem__(self, key, value):
        if key in self.data:
            # Can simply update the proxy
            proxy = self.data[key]  # type: ObjProxy
            proxy.assign(value)
            if proxy.is_saved():
                proxy.save()
        else:
            proxy = self._create_proxy(value)
            self.data[key] = proxy
            if self.is_saved():
                self.save()  # This will cause the proxy to be saved as well

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]
        del self.data[key]
        self._historian.delete(proxy)

    @sync()
    def __iter__(self):
        return self.data.__iter__()

    @sync()
    def __len__(self):
        return len(self.data)

    def _create_proxy(self, value):
        return ObjProxy(value, self._historian)


class LiveRefDict(RefDict):
    """A live dictionary that uses references to refer to contained objects"""
    # pylint: disable=too-few-public-methods
    TYPE_ID = uuid.UUID('16e7e814-8268-46e0-8d8e-6f34132366b9')
    CONTAINER = LiveDict


# endregion


class File(base_savable.BaseSavableObject, metaclass=ABCMeta):
    ATTRS = ('_filename', '_encoding')
    READ_SIZE = 256  # The number of bytes to read at a time

    def __init__(self, filename: str = None, encoding=None):
        super().__init__()
        self._filename = filename
        self._encoding = encoding

    @property
    def filename(self) -> Optional[str]:
        return self._filename

    @property
    def encoding(self) -> Optional[str]:
        return self._encoding

    @abstractmethod
    def open(self, mode='r', **kwargs) -> BinaryIO:
        """Open returning a file like object that supports close() and read()"""

    def from_disk(self, path):
        """Copy the contents of a disk file to this file"""
        with open(str(path), 'r', encoding=self.encoding) as disk_file:
            with self.open('w') as this:
                shutil.copyfileobj(disk_file, this)

    def to_disk(self, folder: [str, Path]):
        """Copy the contents of this file to a file on disk in the given folder"""
        file_path = Path(str(folder)) / self.filename
        with open(str(file_path), 'w', encoding=self._encoding) as disk_file:
            with self.open('r') as this:
                shutil.copyfileobj(this, disk_file)

    def write_text(self, text: str, encoding=None):
        encoding = encoding or self._encoding
        with self.open('w', encoding=encoding) as fileobj:
            fileobj.write(text)

    def read_text(self, encoding=None) -> str:
        """Read the contents of the file as text.
        This function is named as to mirror pathlib.Path"""
        encoding = encoding or self._encoding
        with self.open('r', encoding=encoding) as fileobj:
            return fileobj.read()

    def __str__(self):
        contents = [str(self._filename)]
        if self._encoding is not None:
            contents.append("({})".format(self._encoding))
        return " ".join(contents)

    def __eq__(self, other) -> bool:
        """Compare the contents of two files

        If both files do not exist they are considered equal.
        """
        if not isinstance(other, BaseFile) or self.filename != other.filename:
            return False

        try:
            with self.open() as my_file:
                try:
                    with other.open() as other_file:
                        while True:
                            my_line = my_file.readline(self.READ_SIZE)
                            other_line = other_file.readline(self.READ_SIZE)
                            if my_line != other_line:
                                return False
                            if my_line == '' and other_line == '':
                                return True
                except FileNotFoundError:
                    return False
        except FileNotFoundError:
            # Our file doesn't exist, make sure the other doesn't either
            try:
                with other.open():
                    return False
            except FileNotFoundError:
                return True

    def yield_hashables(self, hasher):
        """Hash the contents of the file"""
        try:
            with self.open('rb') as opened:
                while True:
                    line = opened.read(self.READ_SIZE)
                    if line == b'':
                        return
                    yield line
        except FileNotFoundError:
            yield from hasher.yield_hashables(None)


BaseFile = File

HISTORIAN_TYPES = (Str, List, RefList, LiveList, LiveRefList, Dict, RefDict, LiveDict, LiveRefDict,
                   ObjProxy)



from abc import ABCMeta, abstractmethod
from typing import Type
import uuid

import mincepy
from . import records
from . import types

__all__ = 'TypeHelper', 'WrapperHelper', 'BaseHelper'


def inject_creation_tracking(cls):
    # Check to make sure we don't do this twice!
    if not hasattr(cls, '__orig_new'):
        cls.__orig_new = cls.__new__

        def new(_cls, *_args, **_kwargs):
            inst = cls.__orig_new(_cls)
            hist = mincepy.get_historian()
            if hist is not None:
                hist.created(inst)
            return inst

        cls.__new__ = new


def remove_creation_tracking(cls):
    try:
        cls.__new__ = cls.__orig_new
    except AttributeError:
        pass


class TypeHelper(metaclass=ABCMeta):
    """This interface provides the basic methods necessary to enable a type
    to be compatible with the historian."""
    TYPE = None  # The type this helper corresponds to
    TYPE_ID = None  # The unique id for this type of objects
    IMMUTABLE = False  # If set to true then the object is decoded straight away
    INJECT_CREATION_TRACKING = False

    def __init__(self):
        assert self.TYPE is not None, "Must set the TYPE to a type of or a tuple of types"
        if self.INJECT_CREATION_TRACKING:
            inject_creation_tracking(self.TYPE)

    def new(self, encoded_saved_state):  # pylint: disable=unused-argument
        """Create a new blank object of this type"""
        cls = self.TYPE
        return cls.__new__(cls)

    @abstractmethod
    def yield_hashables(self, obj, hasher):
        """Produce a hash representing the value"""

    @abstractmethod
    def eq(self, one, other) -> bool:  # pylint: disable=invalid-name
        """Determine if two objects are equal"""

    @abstractmethod
    def save_instance_state(self, obj, saver):
        """Save the instance state of an object, should return a saved instance"""

    @abstractmethod
    def load_instance_state(self, obj, saved_state, loader):
        """Take the given blank object and load the instance state into it"""


class BaseHelper(TypeHelper, metaclass=ABCMeta):
    """A base helper that defaults to yielding hashables directly on the object
    and testing for equality using == given two objects.  This behaviour is fairly
    standard and therefor more type helpers will want to subclass from this class."""

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(obj)

    def eq(self, one, other) -> bool:
        return one == other


class WrapperHelper(TypeHelper):
    """Wraps up an object type to perform the necessary Historian actions"""

    # pylint: disable=invalid-name

    def __init__(self, obj_type: Type[types.SavableObject]):
        self.TYPE = obj_type
        self.TYPE_ID = obj_type.TYPE_ID
        super(WrapperHelper, self).__init__()

    def yield_hashables(self, obj, hasher):
        yield from self.TYPE.yield_hashables(obj, hasher)

    def eq(self, one, other) -> bool:
        return self.TYPE.__eq__(one, other)

    def save_instance_state(self, obj: types.Savable, saver):
        return self.TYPE.save_instance_state(obj, saver)

    def load_instance_state(self, obj, saved_state: types.Savable, loader):
        self.TYPE.load_instance_state(obj, saved_state, loader)


class SnapshotRefHelper(TypeHelper):
    """Add ability to store references"""
    TYPE = records.Ref
    TYPE_ID = uuid.UUID('05fe092b-07b3-4ffc-8cf2-cee27aa37e81')

    def eq(self, one, other):
        if not (isinstance(one, records.Ref) and isinstance(other, records.Ref)):
            return False

        return one.obj_id == other.obj_id and one.version == other.version

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(obj.obj_id)
        yield from hasher.yield_hashables(obj.version)

    def save_instance_state(self, obj, saver):
        return obj.to_list()

    def load_instance_state(self, obj, saved_state, loader):
        obj.__init__(*saved_state)



from typing import Type, MutableMapping, Iterable, Any

from . import helpers
from . import types


class TypeRegistry:

    def __init__(self):
        self._helpers = {}  # type: MutableMapping[Type, helpers.TypeHelper]
        self._type_ids = {}  # type: MutableMapping[Any, Type]

    def __contains__(self, item: Type) -> bool:
        return item in self._helpers

    def register_type(
            self, obj_class_or_helper: [helpers.TypeHelper,
                                        Type[types.Object]]) -> helpers.WrapperHelper:
        if isinstance(obj_class_or_helper, helpers.TypeHelper):
            helper = obj_class_or_helper
        else:
            if not issubclass(obj_class_or_helper, types.Object):
                raise TypeError("Type '{}' is nether a TypeHelper nor a SavableObject".format(
                    obj_class_or_helper))
            helper = helpers.WrapperHelper(obj_class_or_helper)

        obj_types = helper.TYPE if isinstance(helper.TYPE, Iterable) else (helper.TYPE,)

        for obj_type in obj_types:
            self._helpers[obj_type] = helper
            self._type_ids[helper.TYPE_ID] = obj_type

        return helper

    def get_type_id(self, obj_type):
        try:
            # Try a direct lookup first
            return self._helpers[obj_type].TYPE_ID
        except KeyError:
            # Try an issubclass lookup as a backup
            for type_id, known_type in self._type_ids.items():
                if issubclass(obj_type, known_type):
                    return type_id
            raise ValueError("Type '{}' is not known".format(obj_type))

    def get_helper(self, type_id_or_type) -> helpers.TypeHelper:
        if isinstance(type_id_or_type, type):
            return self.get_helper_from_obj_type(type_id_or_type)

        return self.get_helper_from_type_id(type_id_or_type)

    def get_helper_from_type_id(self, type_id) -> helpers.TypeHelper:
        try:
            return self.get_helper_from_obj_type(self._type_ids[type_id])
        except KeyError:
            raise TypeError("Type id '{}' not known".format(type_id))

    def get_helper_from_obj_type(self, obj_type: Type) -> helpers.TypeHelper:
        try:
            # Try the direct lookup
            return self._helpers[obj_type]
        except KeyError:
            # Do the full issubclass lookup
            for known_type, helper in self._helpers.items():
                if issubclass(obj_type, known_type):
                    return helper
            raise ValueError("Type '{}' has not been registered".format(obj_type))



"""Classes and function useful for trying out mincepy functionality"""

import uuid

import bson
import mincepy
import pymongo
import pytest


@pytest.fixture
def mongodb_archive():
    client = pymongo.MongoClient()
    db = client.test_database
    mongo_archive = mincepy.mongo.MongoArchive(db)
    yield mongo_archive
    client.drop_database(db)


@pytest.fixture
def historian(mongodb_archive):
    hist = mincepy.Historian(mongodb_archive)
    hist.register_types(mincepy.testing.HISTORIAN_TYPES)
    mincepy.set_historian(hist)
    yield hist
    mincepy.set_historian(None)


class Car(mincepy.BaseSavableObject):
    TYPE_ID = bson.ObjectId('5e075d6244572f823ed93274')
    ATTRS = 'colour', 'make'

    def __init__(self, make='ferrari', colour='red'):
        super(Car, self).__init__()
        self.make = make
        self.colour = colour


class Garage(mincepy.BaseSavableObject):
    TYPE_ID = bson.ObjectId('5e07b40a44572f823ed9327b')
    ATTRS = ('car',)

    def __init__(self, car=None):
        super(Garage, self).__init__()
        self.car = car


class Person(mincepy.BaseSavableObject):
    TYPE_ID = uuid.UUID('d60ca740-9fa6-4002-83f6-e4c91403e41b')
    ATTRS = 'name', 'age'

    def __init__(self, name, age):
        super(Person, self).__init__()
        self.name = name
        self.age = age

    def __eq__(self, other):
        return self.name == other.name and self.age == other.age


class Cycle(mincepy.BaseSavableObject):
    TYPE_ID = uuid.UUID('600fb6ae-684c-4f8e-bed3-47ae06739d29')
    ATTRS = ('_ref',)

    def __init__(self, ref=None):
        super(Cycle, self).__init__()
        self._ref = mincepy.ObjRef(ref)

    @property
    def ref(self):
        return self._ref()

    @ref.setter
    def ref(self, value):
        self._ref = mincepy.ObjRef(value)

    def __eq__(self, other):
        return self.ref is other.ref

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables(id(self.ref))


HISTORIAN_TYPES = Car, Garage, Person, Cycle



"""References module"""
import uuid

from . import exceptions
from . import types
from . import records

__all__ = ('ObjRef',)


class ObjRef(types.SavableObject):
    """A reference to an object instance"""
    TYPE_ID = uuid.UUID('633c7035-64fe-4d87-a91e-3b7abd8a6a28')
    IMMUTABLE = True

    _obj = None
    _ref = None
    _loader = None

    def __init__(self, obj=None, historian=None):
        super().__init__(historian)
        assert not (obj is not None and self._historian.is_primitive(obj)), \
            "Can't create a reference to a primitive type"
        self._obj = obj

    def __str__(self):
        desc = "ref: "
        if self._obj is not None:
            desc += str(self._obj)
        else:
            desc += str(self._ref)
        return desc

    def __call__(self, update=False):
        """Get the object being referenced.  If update is called then the latest version
        will be loaded from the historian"""
        if self._obj is None:
            # This means we were loaded and need to load the object
            if self._ref is None:
                raise RuntimeError("Cannot dereference a None reference")
            # Cache the object
            self._obj = self._loader.load(self._ref)
            assert self._obj is not None, "Loader did not load object"
            self._ref = None
            self._loader = None
        elif update:
            try:
                self._historian.sync(self._obj)
            except exceptions.NotFound:
                pass  # Object must never have been saved and is therefore up to date

        return self._obj

    def __eq__(self, other):
        if not isinstance(other, ObjRef):
            return False

        if self._obj is not None:
            return id(self._obj) == id(other._obj)

        return self._ref == other._ref

    def yield_hashables(self, hasher):
        if self._obj is not None:
            yield from hasher.yield_hashables(id(self._obj))
        else:
            # This will also work if ref is None
            yield from hasher.yield_hashables(self._ref)

    def save_instance_state(self, saver):
        if self._obj is not None:
            ref = saver.ref(self._obj)
        else:
            ref = self._ref

        if ref is not None:
            return ref.to_list()

        return None

    def load_instance_state(self, saved_state, loader):
        super(ObjRef, self).load_instance_state(saved_state, loader)
        # Rely on class default values for members
        if saved_state is not None:
            self._ref = records.Ref(*saved_state)
            self._loader = loader


HISTORIAN_TYPES = (ObjRef,)



from collections import OrderedDict
import typing

import click
import pymongo
from tabulate import tabulate

import mincepy
import mincepy.records


@click.command()
@click.option('--obj-type', default=None, help="The type of object to find")
@click.option('--filter', default=None, help="Filter on the state")
@click.option('--limit', default=0, help="Limit the number of results")
def query(obj_type, filter, limit):
    historian = mincepy.get_historian()

    results = historian.find(obj_type, state=filter, limit=limit, as_objects=False,
                             version=-1)  # type: typing.Sequence[mincepy.DataRecord]

    historian.register_types(mincepy.testing.HISTORIAN_TYPES)

    # Gather by object types
    gathered = {}
    for result in results:
        gathered.setdefault(result.type_id, []).append(result)

    for type_id, records in gathered.items():
        print("type: {}".format(type_id))
        print_records(records, historian)


SCALAR_VALUE = '<value>'
UNSET = ''
REF = 'ref'


def print_records(records: typing.Sequence[mincepy.records.DataRecord], historian):
    columns = OrderedDict()
    refs = []
    for record in records:
        try:
            helper = historian.get_helper(record.type_id)
            type_str = get_type_name(helper.TYPE) + "#{}".format(record.version)
        except KeyError:
            type_str = str(record.get_reference())
        if record.is_deleted_record():
            type_str += " [deleted]"
        refs.append(type_str)
    columns[REF] = refs

    for record in records:
        if not record.is_deleted_record():
            for column_name in get_all_columns(record.state):
                columns[tuple(column_name)] = []

    for column_name in columns.keys():
        if column_name != REF:
            columns[column_name] = [get_value(column_name, record.state) for record in records]

    rows = []
    for row in range(len(records)):
        rows.append([columns[column][row] for column in columns.keys()])

    print(
        tabulate(rows,
                 headers=[
                     ".".join(path) if isinstance(path, tuple) else path for path in columns.keys()
                 ]))


def get_all_columns(state):
    if isinstance(state, dict):
        if 'type_id' in state and 'state' in state:
            yield []
        else:
            for key, value in state.items():
                key_path = [key]
                if isinstance(value, dict):
                    for path in get_all_columns(value):
                        yield key_path + path
                else:
                    yield key_path
    elif isinstance(state, list):
        yield from [(idx,) for idx in range(len(state))]
    else:
        yield SCALAR_VALUE


def get_value(title, state):
    if isinstance(state, (dict, list)):
        idx = title[0]
        value = state[idx]
        # Check for references
        if isinstance(value, dict) and set(value.keys()) == {'type_id', 'state'}:
            return str(mincepy.records.Ref(*value['state']))

        if len(title) > 1:
            return get_value(title[1:], value)

        try:
            return state[idx]
        except (KeyError, IndexError):
            return UNSET

    if title == SCALAR_VALUE:
        return state

    return UNSET


def get_type_name(obj_type):
    try:
        return "{}.{}".format(obj_type.__module__, obj_type.__name__)
    except AttributeError:
        return str(obj_type)


if __name__ == '__main__':
    client = pymongo.MongoClient()
    db = client.test_database
    mongo_archive = mincepy.mongo.MongoArchive(db)

    hist = mincepy.Historian(mongo_archive)
    mincepy.set_historian(hist)

    query()



from . import history

__all__ = 'load', 'save'


def load(*obj_ids_or_refs):
    return history.get_historian().load(*obj_ids_or_refs)


def save(*objs, with_meta=None, return_sref=False):
    return history.get_historian().save(*objs, with_meta=with_meta, return_sref=return_sref)



from argparse import Namespace

import mincepy
import mincepy.records
from mincepy.testing import Car, Cycle


def test_obj_ref_simple(historian: mincepy.Historian):
    a = Cycle()
    a.ref = mincepy.ObjRef(a)
    aid = historian.save(a)
    del a

    loaded = historian.load(aid)
    assert loaded.ref() is loaded


def test_obj_ref_snapshot(historian: mincepy.Historian):
    """Check that a historic snapshot still works with references"""
    ns = Namespace()
    car = Car('honda', 'white')
    ns.car = mincepy.ObjRef(car)
    ns_id_honda = historian.save(ns, return_sref=True)

    car.make = 'fiat'
    ns_id_fiat = historian.save(ns, return_sref=True)
    del ns

    assert ns_id_fiat.version == ns_id_honda.version + 1

    loaded = historian.load(ns_id_honda)
    assert loaded.car().make == 'honda'

    loaded2 = historian.load(ns_id_fiat)
    assert loaded2.car().make == 'fiat'
    assert loaded2.car() is not car

    # Load the 'live' namespace
    loaded3 = historian.load(ns_id_honda.obj_id)
    assert loaded3.car() is car


def test_obj_ref_complex(historian: mincepy.Historian):
    honda = Car('honda')
    nested1 = Namespace()
    nested2 = Namespace()
    parent = Namespace()

    # Both the nested refer to the same car
    nested1.car = mincepy.ObjRef(honda)
    nested2.car = mincepy.ObjRef(honda)

    # Now put them in their containers
    parent.ns1 = mincepy.ObjRef(nested1)
    parent.ns2 = mincepy.ObjRef(nested2)

    parent_id = historian.save(parent)
    del parent

    loaded = historian.load(parent_id)
    assert loaded.ns1() is nested1
    assert loaded.ns2() is nested2
    assert loaded.ns1().car() is loaded.ns2().car()

    fiat = Car('fiat')
    loaded.ns2().car = mincepy.ObjRef(fiat)
    parent_snapshot_id = historian.save(loaded, return_sref=True)
    del loaded

    loaded2 = historian.load_snapshot(mincepy.records.Ref(parent_id, 0))
    assert loaded2.ns1().car().make == 'honda'
    assert loaded2.ns2().car().make == 'honda'
    del loaded2

    loaded3 = historian.load_snapshot(parent_snapshot_id)
    assert loaded3.ns1().car().make == 'honda'
    assert loaded3.ns2().car().make == 'fiat'


def test_null_ref(historian: mincepy.Historian):
    null = mincepy.ObjRef()
    null2 = mincepy.ObjRef()

    assert null == null2

    nid1, nid2 = historian.save(null, null2)
    del null
    loaded = historian.load(nid1)
    assert loaded == null2


def test_ref_load_save_load(historian: mincepy.Historian):
    """This is here to catch a bug that manifested when a reference was saved, loaded and then
    re-saved without being dereferenced in-between.  This would result in the second saved state
    being that of a null reference.  This can only be tested if the reference is stored by value
    as otherwise the historian will not re-save a reference that has not been mutated."""
    ref_list = mincepy.List((mincepy.ObjRef(Car()),))
    assert isinstance(ref_list[0](), Car)

    list_id = ref_list.save()
    del ref_list

    loaded = historian.load(list_id)
    # Re-save
    loaded.save()
    del loaded

    # Re-load
    reloaded = historian.load(list_id)
    # Should still be our car but because of a bug this was a None reference
    assert isinstance(reloaded[0](), Car)



import os
from typing import Optional

from . import archive_factory
from . import historians
from . import plugins

__all__ = 'get_historian', 'set_historian', 'DEFAULT_ARCHIVE_URI', 'ENV_ARCHIVE_URI', 'archive_uri'

DEFAULT_ARCHIVE_URI = 'mongodb://localhost/mincepy'
ENV_ARCHIVE_URI = 'MINCEPY_ARCHIVE'
CURRENT_HISTORIAN = None


def archive_uri():
    """Returns the currently set archive URI to use by default"""
    return os.environ.get(ENV_ARCHIVE_URI, DEFAULT_ARCHIVE_URI)


def create_default_historian():
    uri = archive_uri()
    if uri:
        return archive_factory.historian(uri)

    return None


def get_historian() -> historians.Historian:
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    if CURRENT_HISTORIAN is None:
        # Try creating a new one
        set_historian(create_default_historian())

    return CURRENT_HISTORIAN


def set_historian(new_historian: Optional[historians.Historian], apply_plugins=True):
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    CURRENT_HISTORIAN = new_historian
    if new_historian is not None and apply_plugins:
        new_historian.register_types(plugins.get_types())



import copy
import datetime
import typing
from collections import namedtuple

from . import utils

__all__ = 'OBJ_ID', 'TYPE_ID', 'CREATION_TIME', 'VERSION', 'STATE', 'SNAPSHOT_TIME', \
          'SNAPSHOT_HASH', 'EXTRAS', 'ExtraKeys', 'DELETED', 'DataRecord', 'Ref', \
          'DataRecordBuilder'

OBJ_ID = 'obj_id'
TYPE_ID = 'type_id'
CREATION_TIME = 'creation_time'
VERSION = 'version'
STATE = 'state'
STATE_TYPES = 'state_types'
SNAPSHOT_HASH = 'snapshot_hash'
SNAPSHOT_TIME = 'snapshot_time'
EXTRAS = 'extras'


class ExtraKeys:
    CREATED_BY = '_created_by'  # The ID of the process the data was created in
    COPIED_FROM = '_copied_from'  # The reference to the snapshot that this object was copied from
    USER = '_user'  # The user that saved this snapshot
    HOSTNAME = '_hostname'  # The hostname of the computer this snapshot was saved on


DELETED = '!!deleted'  # Special state to denote a deleted record

IdT = typing.TypeVar('IdT')  # The archive ID type


class Ref(typing.Generic[IdT]):
    """Reference information for looking up a snapshot"""

    def __init__(self, obj_id, version):
        super(Ref, self).__init__()
        self._obj_id = obj_id
        self._version = version

    def __str__(self):
        return "{}#{}".format(self._obj_id, self._version)

    def __hash__(self):
        return (self.obj_id, self.version).__hash__()

    def __eq__(self, other):
        if not isinstance(other, Ref):
            return False

        return self.obj_id == other.obj_id and self.version == other.version

    @property
    def obj_id(self) -> IdT:
        return self._obj_id

    @property
    def version(self):
        return self._version

    def to_list(self) -> list:
        """Convenience to represent the reference as a list"""
        return [self.obj_id, self.version]


class DataRecord(
        namedtuple(
            'DataRecord',
            (
                # Object properties
                OBJ_ID,  # The ID of the object (spanning all snapshots)
                TYPE_ID,  # The type ID of this object
                CREATION_TIME,  # The time this object was created
                # Snapshot properties
                VERSION,  # The ID of this particular snapshot of the object
                STATE,  # The saved state of the object
                STATE_TYPES,  # The historian types saved in the state
                SNAPSHOT_HASH,  # The hash of the state
                SNAPSHOT_TIME,  # The time this snapshot was created
                EXTRAS,  # Additional data stored with the snapshot
            ))):
    """An immutable record that describes a snapshot of an object"""

    @classmethod
    def defaults(cls) -> dict:
        """Returns a dictionary of default values, the caller owns the dict and is free to modify it"""
        return {
            CREATION_TIME: None,
            SNAPSHOT_TIME: None,
            EXTRAS: {},
        }

    @classmethod
    def new_builder(cls, **kwargs) -> utils.NamedTupleBuilder:
        """Get a builder for a new data record, the version will be set to 0"""
        values = cls.defaults()
        values.update({
            CREATION_TIME: utils.DefaultFromCall(datetime.datetime.now),
            VERSION: 0,
            SNAPSHOT_TIME: utils.DefaultFromCall(datetime.datetime.now),
        })
        values.update(kwargs)
        return utils.NamedTupleBuilder(cls, values)

    @property
    def created_by(self):
        """Convenience property to get the creator from the extras"""
        return self.get_extra(ExtraKeys.CREATED_BY)

    def is_deleted_record(self):
        """Does this record represent the object having been deleted"""
        return self.state == DELETED

    def get_reference(self) -> Ref:
        """Get a reference for this data record"""
        return Ref(self.obj_id, self.version)

    def get_copied_from(self) -> Ref:
        """Get the reference of the data record this object was originally copied from"""
        obj_ref = self.get_extra(ExtraKeys.COPIED_FROM)
        if obj_ref is None:
            return None

        return Ref(*obj_ref)

    def get_extra(self, name):
        """Convenience function to get an extra from the record, returns None if the extra doesn't exist"""
        return self.extras.get(name, None)

    def copy_builder(self, **kwargs) -> utils.NamedTupleBuilder:
        """Get a copy builder from this DataRecord instance.  The following attributes will be copied over:
            * type_id
            * state [deepcopy]
            * snapshot_hash
        the version will be set to 0 and the creation time to now
        """
        defaults = self.defaults()
        defaults.update({
            TYPE_ID: self.type_id,
            CREATION_TIME: utils.DefaultFromCall(datetime.datetime.now),
            STATE: copy.deepcopy(self.state),
            STATE_TYPES: copy.deepcopy(self.state_types),
            SNAPSHOT_HASH: self.snapshot_hash,
            VERSION: 0,
            SNAPSHOT_TIME: utils.DefaultFromCall(datetime.datetime.now),
        })
        defaults[EXTRAS][ExtraKeys.COPIED_FROM] = self.get_reference().to_list()
        defaults.update(kwargs)
        return utils.NamedTupleBuilder(type(self), defaults)

    def child_builder(self, **kwargs) -> utils.NamedTupleBuilder:
        """
        Get a child builder from this DataRecord instance.  The following attributes will be copied over:
            * obj_id
            * type_id
            * creation_time
            * created_by
        and version will be incremented by one.
        """
        defaults = self.defaults()
        defaults.update({
            OBJ_ID: self.obj_id,
            TYPE_ID: self.type_id,
            CREATION_TIME: self.creation_time,
            VERSION: self.version + 1,
            SNAPSHOT_TIME: utils.DefaultFromCall(datetime.datetime.now),
            EXTRAS: self.extras.copy(),
        })
        defaults.update(kwargs)
        return utils.NamedTupleBuilder(type(self), defaults)


DataRecordBuilder = utils.NamedTupleBuilder[DataRecord]


def make_deleted_builder(last_record: DataRecord) -> DataRecordBuilder:
    """Get a record that represents the deletion of this object"""
    return last_record.child_builder(state=DELETED, state_types=None, snapshot_hash=None)



import uuid

import mincepy.records
from . import archives


class InMemory(archives.BaseArchive):
    """An archive that keeps things in memory"""

    def __init__(self, codecs=tuple()):
        super(InMemory, self).__init__(codecs)
        self._records = {}

    def create_archive_id(self):
        return uuid.uuid4()

    def save(self, record: mincepy.records.DataRecord):
        self._records[record.obj_id] = record

    def load(self, snapshot_id) -> mincepy.records.DataRecord:
        return self._records[snapshot_id]



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = "Martin Uhrin"
__license__ = "GPLv3 and MIT, see LICENSE file"

about = {}
with open('mincepy/version.py') as f:
    exec(f.read(), about)

setup(name='mincepy',
      version=about['__version__'],
      description="Object storage with versioning made simple",
      long_description=open('README.rst').read(),
      url='https://github.com/muhrin/mincepy.git',
      author='Martin Uhrin',
      author_email='martin.uhrin.10@ucl.ac.uk',
      license=__license__,
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='database schemaless nosql object-store concurrent optimistic-locking',
      install_requires=[
          'pymongo',
          'bidict',
          'pyblake2; python_version<"3.6"',
          'pytray>=0.2.1',
          'stevedore',
          'click',
      ],
      extras_require={
          'cli': ['click', 'tabulate'],
          'gui': ['mincepy-gui'],
          'dev': [
              'pip',
              'pytest>4',
              'pytest-cov',
              'pre-commit',
              'yapf',
              'prospector',
              'pylint',
              'twine',
          ],
      },
      packages=['mincepy', 'mincepy.cli'],
      include_package_data=True,
      test_suite='test',
      provides=['mincepy.plugins'],
      entry_points={
          'console_scripts': ['mince = mincepy.cli.main:mince'],
          'mincepy.plugins.types': ['native_types = mincepy.provides:get_types', ]
      })



import mincepy.records
from mincepy.testing import *


def test_find_state(historian: mincepy.Historian):
    honda = Car('honda', 'green')
    porsche = Car('porsche', 'black')
    red_honda = Car('honda', 'red')
    fiat = Car('fiat', 'green')

    historian.save(honda, porsche, red_honda, fiat)
    hondas = list(historian.find(Car, state={'make': 'honda'}))
    assert len(hondas) == 2
    assert honda in hondas
    assert red_honda in hondas

    # Try without type
    greens = list(historian.find(state={'colour': 'green'}))
    assert len(greens) == 2
    assert honda in greens
    assert fiat in greens


def test_find_skip(historian: mincepy.Historian):
    cars = []
    for idx in range(10):
        cars.append(Car(idx))

    historian.save(*cars)
    # Try live
    makes = set(range(10))
    # Skip two at a time limiting to two at a time, effectively paginating
    for skip in range(5):
        results = list(historian.find(obj_type=Car, limit=2, skip=skip * 2))
        assert len(results) == 2, "Got no results on page {}".format(skip)
        makes.remove(results[0].make)
        makes.remove(results[1].make)

    assert not makes


def test_simple_sort(historian: mincepy.Historian):
    cars = mincepy.builtins.RefList()
    for idx in range(10):
        cars.append(Car(idx))

    historian.save(cars)
    results = list(historian.find(Car, sort=mincepy.records.CREATION_TIME, as_objects=False))
    for idx, result in enumerate(results[1:]):
        # No need to subtract 1 from idx as we're already one behind because of the slicing
        assert result.creation_time >= results[idx].creation_time



import functools
import uuid

from . import base_savable
from . import builtins
from . import history
from . import refs

__all__ = ('track', 'FunctionCall')


class InvalidStateError(Exception):
    pass


class FunctionCall(base_savable.BaseSavableObject):
    TYPE_ID = uuid.UUID('dcacc483-c650-432e-b835-122f78e7a758')
    ATTRS = ('_function', '_args', '_kwargs', '_result', '_exception', '_done')

    def __init__(self, func, *args, **kwargs):
        super(FunctionCall, self).__init__()
        self._function = func.__name__
        self._args = builtins.RefList(args)
        self._kwargs = builtins.RefDict(kwargs)
        self._result = None
        self._exception = None
        self._done = False

    @property
    def function(self) -> str:
        return self._function

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def result(self):
        if not self.done():
            raise InvalidStateError("Not done yet")

        return self._result()

    def set_result(self, result):
        assert not self._done
        self._result = refs.ObjRef(result)
        self._done = True

    def set_exception(self, exc):
        assert not self._done
        self._exception = str(exc)
        self._done = True

    def done(self):
        return self._done


def track(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        hist = history.get_historian()
        call = FunctionCall(func, *args, **kwargs)
        try:
            result = func(*args, **kwargs)
            call.set_result(result)
        except Exception as exc:  # pylint: disable=broad-except
            call.set_exception(exc)
        else:
            return result
        finally:
            hist.save(call)

    return wrapper



import pymongo
import pymongo.uri_parser

from . import historians
from . import mongo

__all__ = 'archive', 'historian'

from . import plugins


def archive(uri: str):
    """Create an archive type based on a uri string"""
    if uri.startswith('mongodb'):
        # Format is:
        # mongodb://[username:password@]host1[:port1][,...hostN[:portN]][/[database][?options]]
        parsed = pymongo.uri_parser.parse_uri(uri)
        if not parsed.get('database', None):
            raise ValueError("Failed to supply database on MongoDB uri: {}".format(uri))
        client = pymongo.MongoClient(uri)
        database = client[parsed['database']]
        return mongo.MongoArchive(database)

    raise ValueError("Unknown archive string: {}".format(uri))


def historian(archive_uri: str, apply_plugins=True):
    """Convenience function to create a standard historian directly from an archive URI"""
    historian = historians.Historian(archive(archive_uri))
    if apply_plugins:
        historian.register_types(plugins.get_types())

    return historian



"""Tools to analyse the state of the historian and archive"""

from typing import Iterable

import mincepy

__all__ = 'get_type_name', 'get_table'

SCALAR_VALUE = '[value]'
UNSET = '[unset]'


def get_type_name(obj_type) -> str:
    """Get a simplified type name for the type, say, given by a helper"""
    try:
        return "{}.{}".format(obj_type.__module__, obj_type.__name__)
    except AttributeError:
        return str(obj_type)


def get_table(records: Iterable, progress_callback=None):
    dicts = []
    obj_ids = []

    for idx, record in enumerate(records):
        # First get all the columns in this record
        paths = list(get_paths(record.state))
        state_values = {}

        for path in paths:
            if isinstance(path, list):
                str_path = ".".join(path)
            else:
                str_path = path
            value = get_value(path, record.state)
            state_values[str_path] = str(value)

        obj_ids.append(record.obj_id)
        dicts.append(state_values)
        if progress_callback is not None:
            progress_callback(idx)

    # Now we've collected all the information we need let's gather it all together
    # Get all the unique keys
    all_keys = set()
    for state_values in dicts:
        all_keys.update(state_values.keys())

    # 0th row always contains column names
    rows = [['Obj id'] + list(all_keys)]
    # Now fill in all the values
    for obj_id, state_values in zip(obj_ids, dicts):
        row = [str(obj_id)]
        for key in rows[0][1:]:
            row.append(state_values.get(key, UNSET))
        rows.append(row)

    return rows


def get_paths(state):
    if isinstance(state, dict):
        for key, value in state.items():
            key_path = [key]
            if isinstance(value, dict):
                for path in get_paths(value):
                    yield key_path + path
            else:
                yield key_path
    elif isinstance(state, list):
        yield from [(idx,) for idx in range(len(state))]
    else:
        yield SCALAR_VALUE


def get_value(path, state):
    if isinstance(state, (dict, list)):
        idx = path[0]
        value = state[idx]
        # Check for references
        if isinstance(value, dict) and set(value.keys()) == {'type_id', 'state'}:
            return str(mincepy.Ref(*value['state']))

        if len(path) > 1:
            return get_value(path[1:], value)

        try:
            return state[idx]
        except (KeyError, IndexError):
            return UNSET

    if path == SCALAR_VALUE:
        return state

    return UNSET



""""Tests of metadata storage"""
import pytest

import mincepy
from mincepy.testing import Car


def test_metadata_simple(historian: mincepy.Historian):
    car = Car()
    ferrari_id = historian.save(car, with_meta={'reg': 'VD395'})
    # Check that we get back what we just set
    assert historian.get_meta(ferrari_id) == {'reg': 'VD395'}

    car.make = 'fiat'
    red_fiat_id = historian.save(car)
    # Check that the metadata is shared
    assert historian.get_meta(red_fiat_id) == {'reg': 'VD395'}

    historian.set_meta(ferrari_id, {'reg': 'N317'})
    # Check that this saves the metadata on the object level i.e. both are changed
    assert historian.get_meta(ferrari_id) == {'reg': 'N317'}
    assert historian.get_meta(red_fiat_id) == {'reg': 'N317'}


def test_metadata_using_object_instance(historian: mincepy.Historian):
    car = Car()
    historian.save(car, with_meta={'reg': 'VD395'})
    # Check that we get back what we just set
    assert historian.get_meta(car) == {'reg': 'VD395'}

    car.make = 'fiat'
    historian.save(car)
    # Check that the metadata is shared
    assert historian.get_meta(car) == {'reg': 'VD395'}

    historian.set_meta(car, {'reg': 'N317'})
    assert historian.get_meta(car) == {'reg': 'N317'}


def test_metadata_multiple(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    zonda = Car('zonda', 'yellow')

    historian.save(honda, zonda, with_meta=({'reg': 'H123'}, {'reg': 'Z456'}))

    assert historian.get_meta(honda) == {'reg': 'H123'}
    assert historian.get_meta(zonda) == {'reg': 'Z456'}


def test_metadata_wrong_type(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    with pytest.raises(TypeError):
        historian.save(honda, with_meta=['a', 'b'])


def test_metadata_update(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    historian.save(honda, with_meta={'reg': 'H123', 'vin': 1234, 'owner': 'Mike'})
    historian.update_meta(honda, {'vin': 5678, 'owner': 'Mart'})
    assert historian.get_meta(honda) == {'reg': 'H123', 'vin': 5678, 'owner': 'Mart'}


def test_metadata_update_inexistant(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    historian.save(honda)
    # If the data doesn't exist in the metadata already we expect an update to simply insert
    historian.update_meta(honda, {'reg': 'H123', 'vin': 1234})
    assert historian.get_meta(honda) == {'reg': 'H123', 'vin': 1234}


def test_metadata_find(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    honda2 = Car('honda', 'white')
    historian.save(honda, with_meta={'reg': 'H123', 'vin': 1234})
    historian.save(honda2)

    results = list(historian.find(Car, meta={'reg': 'H123'}))
    assert len(results) == 1
